from unittest import TestCase

from clayrs.content_analyzer.information_processor.spacy import Spacy


class TestSpacy(TestCase):

    def test_process_stopwords(self):
        spa = Spacy(stopwords_removal=True)
        expected = ["striped", "bats", "hanging", "feet", "best"]
        result = spa.process("The striped bats are hanging on their feet for the best")
        self.assertEqual(expected, result)

        # add new stopwords
        spa = Spacy(stopwords_removal=True, new_stopwords=['bats', 'best'])
        expected = ["striped", "hanging", "feet"]
        result = spa.process("The striped bats are hanging on their feet for the best")
        self.assertEqual(expected, result)

        # allow some stopwords
        spa = Spacy(stopwords_removal=True, not_stopwords=['The', 'the', 'on'])
        expected = ["The", "striped", "bats", "hanging", "on", "feet", "the", "best"]
        result = spa.process("The striped bats are hanging on their feet for the best")
        self.assertEqual(expected, result)

        # add new stopwords and allow some stopwords
        spa = Spacy(stopwords_removal=True, new_stopwords=['bats', 'best'], not_stopwords=['The', 'the', 'on'])
        expected = ["The", "striped", "hanging", "on", "feet", "the"]
        result = spa.process("The striped bats are hanging on their feet for the best")
        self.assertEqual(expected, result)

    def test_process_lemmatization(self):
        spa = Spacy(lemmatization=True)
        expected = ["the", "stripe", "bat", "be", "hang", "on", "their", "foot", "for", "good"]
        result = spa.process("The striped bats are hanging on their feet for best")
        self.assertEqual(expected, result)

    def test_remove_punctuation(self):
        spa = Spacy(remove_punctuation=True)
        expected = ["Hello", "there", "How", "are", "you", "I", "'m", "fine", "thanks"]
        result = spa.process("Hello there. How are you? I'm fine, thanks.")
        self.assertEqual(expected, result)

    def test_url_tagging(self):
        spa = Spacy(url_tagging=True)
        expected = ["The", "striped", "<URL>", "bats", "<URL>", "are", "<URL>", "hanging", "on", "their", "feet",
                    "for", "best", "<URL>"]
        result = spa.process("The striped http://facebook.com bats https://github.com are "
                             "http://facebook.com hanging on their feet for best http://twitter.it")
        self.assertEqual(expected, result)

    def test_entity_recognition(self):
        spa = Spacy(named_entity_recognition=True)
        expected = ["Facebook", "was", "fined", "by", "<Hewlett_ORG_B>", "<Packard_ORG_I>", "for", "spending",
                    "<100_MONEY_B>", "<€_MONEY_I>", "to", "buy", "<Cristiano_PERSON_B>", "<Ronaldo_PERSON_I>",
                    "from", "<Juventus_ORG_B>"]
        result = spa.process(
            "Facebook was fined by Hewlett Packard for spending 100€ to buy Cristiano Ronaldo from "
            "Juventus")

        self.assertEqual(expected, result)

    def test_multiple_operations(self):
        # Test for lemmatization with multiple whitespaces removal
        spa = Spacy(strip_multiple_whitespaces=True, lemmatization=True)
        expected = ["the", "stripe", "bat", "be", "hang", "on", "their", "foot", "for", "good"]
        result = spa.process("The   striped  bats    are    hanging   on   their    feet   for  best")
        self.assertEqual(expected, result)

        # Test for lemmatization with multiple whitespaces removal and URL tagging
        spa = Spacy(strip_multiple_whitespaces=True, lemmatization=True, url_tagging=True)
        expected = ["the", "stripe", "<URL>", "bat", "<URL>", "be", "<URL>", "hang", "on", "their", "foot", "for",
                    "good", "<URL>"]
        result = spa.process(
            "The   striped http://facebook.com bats https://github.com   are   http://facebook.com "
            "hanging   on   their    feet   for  best  http://twitter.it")

        self.assertEqual(expected, result)

        # Test for lemmatization, multiple whitespaces removal, URL tagging and stopwords removal with new stopwords
        spa = Spacy(lemmatization=True, strip_multiple_whitespaces=True, url_tagging=True, stopwords_removal=True,
                    new_stopwords=['feet'])
        expected = ["stripe", "<URL>", "bat", "<URL>", "<URL>", "hang", "good", "<URL>"]
        result = spa.process(
            "The   striped http://facebook.com bats https://github.com   are   http://facebook.com "
            "hanging   on   their    feet   for  best  http://twitter.it")
        self.assertEqual(expected, result)

        # Test for lemmatization, multiple whitespaces removal, URL tagging, stemming, stopwords removal with new
        # stopwords and remove punctuation
        spa = Spacy(lemmatization=True, strip_multiple_whitespaces=True, url_tagging=True, stopwords_removal=True,
                    new_stopwords=['feet'], remove_punctuation=True)
        expected = ["stripe", "<URL>", "bat", "<URL>", "<URL>", "hang", "good", "<URL>"]
        result = spa.process(
            "The   striped http://facebook.com bats https://github.com   are.   http://facebook.com hanging   On   "
            "their.    feet;   for:  best  http://twitter.it")

        self.assertEqual(expected, result)



import inspect
import itertools
from typing import List, Dict, Callable
import warnings

from ekphrasis.classes.tokenizer import SocialTokenizer
from ekphrasis.classes.segmenter import Segmenter

from ekphrasis.classes.preprocessor import TextPreProcessor
from ekphrasis.classes.spellcorrect import SpellCorrector

from clayrs.content_analyzer.information_processor.information_processor import NLP
from clayrs.utils.automatic_methods import autorepr

with warnings.catch_warnings():
    warnings.simplefilter(action='ignore', category=FutureWarning)
    social_tokenizer_ekphrasis = SocialTokenizer(lowercase=True).tokenize


class Ekphrasis(NLP):
    """
    Interface to the Ekphrasis library for natural language processing features

    Examples:

        * Normalize email and percentage tokens but omit email ones:
        >>> ek = Ekphrasis(omit=['email'], normalize=['email', 'percent'])
        >>> ek.process("this is an email: alias@mail.com and this is a percent 23%")
        ['this', 'is', 'an', 'email', ':', 'and', 'this', 'is', 'a', 'percent', '<percent>']

        * Unpack contractions on running text:
        >>> ek = Ekphrasis(unpack_contractions=True)
        >>> ek.process("I can't do this because I won't and I shouldn't")
        ['i', 'can', 'not', 'do', 'this', 'because', 'i', 'will', 'not', 'and', 'i', 'should', 'not']

        * Unpack hashtag using statistics from 'twitter' corpus:
        >>> ek = Ekphrasis(unpack_hashtags=True, segmenter='twitter')
        >>> ek.process("#next #gamedev #retrogaming #coolphoto no unpack")
        ['next', 'game', 'dev', 'retro', 'gaming', 'cool', 'photo', 'no', 'unpack']

        * Annotate words in CAPS and repeated tokens with single tag for CAPS words:
        >>> ek = Ekphrasis(annotate=['allcaps', 'repeated'], all_caps_tag='single')
        >>> ek.process("this is good !!! text and a SHOUTED one")
        ['this', 'is', 'good', '!', '<repeated>', 'text', 'and', 'a', 'shouted', '<allcaps>', 'one']

        * Perform segmentation using statistics from 'twitter' corpus:
        >>> ek = Ekphrasis(segmentation=True, segmenter='twitter')
        >>> ek.process("thewatercooler exponentialbackoff no segmentation")
        ['the', 'watercooler', 'exponential', 'back', 'off', 'no', 'segmentation']

        * Substitute words with custom tokens:
        >>> ek = Ekphrasis(dicts=[{':)': '<happy>', ':(': '<sad>'}])
        >>> ek.process("Hello :) how are you? :(")
        ['Hello', '<happy>', 'how', 'are', 'you', '?', '<sad>']

        * Perform spell correction on text and on elongated words by using statistics from default 'english' corpus:
        >>> Ekphrasis(spell_correction=True, spell_correct_elong=True)
        >>> ek.process("This is huuuuge. The korrect way of doing tihngs is not the followingt")
        ["this", 'is', 'huge', '.', 'the', 'correct', "way", "of", "doing", "things", "is", 'not', 'the',
        'following']

    Args:
        omit: Choose what tokens that you want to omit from the text.

            Possible values: ***['email', 'percent', 'money', 'phone', 'user','time', 'url', 'date', 'hashtag']***

            Important Notes:

                1 - the token in this list must be present in the `normalize`
                    list to have any effect!
                2 - put url at front, if you plan to use it.
                    Messes with the regexes!
                3 - if you use hashtag then unpack_hashtags will
                    automatically be set to False

        normalize: Choose what tokens that you want to normalize from the text.
            Possible values: ***['email', 'percent', 'money', 'phone', 'user', 'time', 'url', 'date', 'hashtag']***

            For example: myaddress@mysite.com -> `<email>`

            Important Notes:

                1 - put url at front, if you plan to use it.
                    Messes with the regexes!
                2 - if you use hashtag then unpack_hashtags will
                    automatically be set to False

        unpack_contractions: Replace *English* contractions in running text with their unshortened forms

            for example: can't -> can not, wouldn't -> would not, and so on...

        unpack_hashtags: split a hashtag to its constituent words.

            for example: #ilikedogs -> i like dogs

        annotate: add special tags to special tokens.

            Possible values: ['hashtag', 'allcaps', 'elongated', 'repeated']

            for example: myaddress@mysite.com -> myaddress@mysite.com <email>

        corrector: define the statistics of what corpus you would like to use [english, twitter].
            Be sure to set `spell_correction` to True if you want to perform
            spell correction on the running text

        tokenizer: callable function that accepts a string and returns a list of strings.
            If no tokenizer is provided then the text will be tokenized on whitespace

        segmenter: define the statistics of what corpus you would like to use [english, twitter].
            Be sure to set `segmentation` to True if you want to perform segmentation on the running text

        all_caps_tag: how to wrap the capitalized words
            Note: applicable only when `allcaps` is included in the `annotate` list
            Possible values ***[single, wrap, every]***:

                - single: add a tag after the last capitalized word
                    for example: "SHOUTED TEXT" -> "shouted text <allcaps>"
                - wrap: wrap all words with opening and closing tags
                    for example: "SHOUTED TEXT" -> "<allcaps> shouted text </allcaps>"
                - every: add a tag after each word
                    for example: "SHOUTED TEXT" -> "shouted <allcaps> text <allcaps>"

        spell_correction: If set to True, running text will be spell corrected using statistics of corpus set in
            `corrector` parameter

        segmentation: If set to True, running text will be segmented using statistics of corpus set in
            `corrector` parameter

            for example: exponentialbackoff -> exponential back off

        spell_correct_elong: choose if you want to perform spell correction after the normalization of elongated words.

            *significantly affects performance (speed)*

        spell_correction: choose if you want to perform spell correction to the text.

            *significantly affects performance (speed)*
    """

    def __init__(self, *,
                 omit: List = None,
                 normalize: List = None,
                 unpack_contractions: bool = False,
                 unpack_hashtags: bool = False,
                 annotate: List = None,
                 corrector: str = None,
                 tokenizer: Callable = social_tokenizer_ekphrasis,
                 segmenter: str = None,
                 all_caps_tag: str = None,
                 spell_correction: bool = False,
                 segmentation: bool = False,
                 dicts: List[Dict] = None,
                 spell_correct_elong: bool = False):

        # ekphrasis has default values for arguments not passed. So if they are not evaluated in our class,
        # we simply don't pass them to ekphrasis
        kwargs_to_pass = {argument: arg_value for argument, arg_value in zip(locals().keys(), locals().values())
                          if argument != 'self' and arg_value is not None}

        self.text_processor = TextPreProcessor(**kwargs_to_pass)

        self.spell_correct_elong = spell_correct_elong

        self.sc = None
        if spell_correction is True:
            if corrector is not None:
                self.sc = SpellCorrector(corpus=corrector)
            else:
                self.sc = SpellCorrector()

        self.segmentation = segmentation
        self.ws = None
        if segmentation is True:
            if segmenter is not None:
                self.ws = Segmenter(corpus=segmenter)
            else:
                self.ws = Segmenter()

        self._repr_string = autorepr(self, inspect.currentframe())

    def __spell_check(self, field_data):
        """
        Correct any spelling errors
        Args:
            field_data: text to correct
        Returns:
            field_data: correct text
        """

        def correct_word(word):
            if self.spell_correct_elong:
                # normalize to at most 2 repeating chars
                word = self.text_processor.regexes["normalize_elong"].sub(r'\1\1', word)

                normalized = self.sc.normalize_elongated(word)
                if normalized:
                    word = normalized

            return self.sc.correct_word(word, fast=True)

        return [correct_word(word) for word in field_data]

    def __word_segmenter(self, field_data) -> List[str]:
        """
        Split words together
        Args:
            field_data: Text to be processed
        Returns (List[str): Text with splitted words
        """
        word_seg_list = [self.ws.segment(word) for word in field_data]

        word_seg_list = itertools.chain.from_iterable([word.split() for word in word_seg_list])

        return list(word_seg_list)

    def process(self, field_data: str) -> List[str]:
        """
        Args:
            field_data: Running text to be processed
        Returns:
            field_data: List of str representing running text preprocessed
        """
        field_data = self.text_processor.pre_process_doc(field_data)
        if self.sc is not None:
            field_data = self.__spell_check(field_data)
        if self.ws is not None:
            field_data = self.__word_segmenter(field_data)
        return field_data

    def __eq__(self, other):
        if isinstance(other, Ekphrasis):
            return self.text_processor.omit == other.text_processor.omit and \
                   self.text_processor.backoff == other.text_processor.backoff and \
                   self.text_processor.unpack_contractions == other.text_processor.unpack_contractions and \
                   self.text_processor.include_tags == other.text_processor.include_tags and \
                   self.text_processor.corrector_corpus == other.text_processor.corrector_corpus and \
                   self.text_processor.tokenizer == other.text_processor.tokenizer and \
                   self.text_processor.segmenter_corpus == other.text_processor.segmenter_corpus and \
                   self.text_processor.all_caps_tag == other.text_processor.all_caps_tag and \
                   self.text_processor.spell_correction == other.text_processor.spell_correction and \
                   self.segmentation == other.segmentation and \
                   self.text_processor.dicts == other.text_processor.dicts and \
                   self.spell_correct_elong == other.spell_correct_elong
        return False

    def __str__(self):
        return "Ekphrasis"

    def __repr__(self):
        return self._repr_string



import os
import shutil
import unittest
from unittest import TestCase

import pandas as pd

from clayrs.content_analyzer.ratings_manager.score_processor import NumberNormalizer
from clayrs.content_analyzer.ratings_manager.ratings import Ratings, Interaction
from clayrs.content_analyzer.raw_information_source import JSONFile
from test import dir_test_files

file_path = os.path.join(dir_test_files, 'test_import_ratings.json')
raw_source = JSONFile(file_path)
raw_source_content = list(raw_source)


class TestInteraction(TestCase):
    def test_all(self):
        interaction_no_timestamp = Interaction('u1', 'i1', 5)

        self.assertTrue(interaction_no_timestamp.user_id == 'u1')
        self.assertTrue(interaction_no_timestamp.item_id == 'i1')
        self.assertTrue(interaction_no_timestamp.score == 5)
        self.assertIsNone(interaction_no_timestamp.timestamp)

        interaction_w_timestamp = Interaction('u2', 'i2', 4, 'timestamp')

        self.assertTrue(interaction_w_timestamp.user_id == 'u2')
        self.assertTrue(interaction_w_timestamp.item_id == 'i2')
        self.assertTrue(interaction_w_timestamp.score == 4)
        self.assertTrue(interaction_w_timestamp.timestamp == 'timestamp')

    def test__eq(self):

        interaction1_same_no_timestamp = Interaction('u1', 'i1', 5)
        interaction2_same_no_timestamp = Interaction('u1', 'i1', 5)
        interaction3_different_no_timestamp = Interaction('u2', 'i1', 2)

        self.assertEqual(interaction1_same_no_timestamp, interaction2_same_no_timestamp)
        self.assertNotEqual(interaction1_same_no_timestamp, interaction3_different_no_timestamp)

        interaction1_same_w_timestamp = Interaction('u1', 'i1', 5, '112222')
        interaction2_same_w_timestamp = Interaction('u1', 'i1', 5, '112222')
        interaction3_different_w_timestamp = Interaction('u2', 'i1', 2, '223333')

        self.assertEqual(interaction1_same_w_timestamp, interaction2_same_w_timestamp)
        self.assertNotEqual(interaction1_same_w_timestamp, interaction3_different_w_timestamp)
        self.assertNotEqual(interaction1_same_no_timestamp, interaction1_same_w_timestamp)

    def test__hash(self):

        interaction1_same_no_timestamp = Interaction('u1', 'i1', 5)
        interaction2_same_no_timestamp = Interaction('u1', 'i1', 5)
        interaction3_same_w_timestamp = Interaction('u2', 'i1', 2, '11222')

        self.assertEqual(hash(interaction1_same_no_timestamp), hash(interaction2_same_no_timestamp))
        self.assertNotEqual(hash(interaction1_same_no_timestamp), hash(interaction3_same_w_timestamp))


class TestRatings(TestCase):

    # this only tests user, item and score column
    def _check_equal_expected_result(self, rat: Ratings):
        self.assertTrue(len(rat) != 0)

        user_id_expected = [str(row['user_id']) for row in raw_source_content]
        item_id_expected = [str(row['item_id']) for row in raw_source_content]
        score_expected = [float(row['stars']) for row in raw_source_content]

        expected_ratings = [(user, item, score) for user, item, score in zip(user_id_expected,
                                                                             item_id_expected,
                                                                             score_expected)]

        user_id_result = rat.user_id_column
        item_id_result = rat.item_id_column
        score_result = rat.score_column

        result_ratings = [(user, item, score) for user, item, score in zip(user_id_result,
                                                                           item_id_result,
                                                                           score_result)]

        self.assertTrue(all(isinstance(user_id, str) for user_id in user_id_result))
        self.assertTrue(all(isinstance(item_id, str) for item_id in item_id_result))
        self.assertTrue(all(isinstance(score, float) for score in score_result))

        self.assertCountEqual(expected_ratings, result_ratings)

    def test_import_ratings_by_key(self):
        rat = Ratings(
            source=raw_source,
            user_id_column='user_id',
            item_id_column='item_id',
            score_column='stars')

        self._check_equal_expected_result(rat)

        self.assertTrue(len(rat.timestamp_column) == 0)

    def test_import_ratings_by_index(self):
        rat = Ratings(
            source=raw_source,
            user_id_column=0,
            item_id_column=1,
            score_column=4)

        self._check_equal_expected_result(rat)

        self.assertTrue(len(rat.timestamp_column) == 0)

    def test_import_ratings_w_timestamp_key(self):
        rat = Ratings(
            source=raw_source,
            user_id_column=0,
            item_id_column=1,
            score_column=4,
            timestamp_column="timestamp"
        )

        self._check_equal_expected_result(rat)

        timestamp_expected = [row['timestamp'] for row in raw_source_content]

        timestamp_result = rat.timestamp_column

        self.assertTrue(all(isinstance(timestamp, str) for timestamp in timestamp_result))
        self.assertEqual(timestamp_expected, timestamp_result)

    def test_import_ratings_w_timestamp_index(self):
        rat = Ratings(
            source=raw_source,
            user_id_column=0,
            item_id_column=1,
            score_column=4,
            timestamp_column=5
        )

        self._check_equal_expected_result(rat)

        timestamp_expected = [row['timestamp'] for row in raw_source_content]

        timestamp_result = rat.timestamp_column

        self.assertTrue(all(isinstance(timestamp, str) for timestamp in timestamp_result))
        self.assertEqual(timestamp_expected, timestamp_result)

    def test_import_ratings_w_score_processor(self):
        rat = Ratings(
            source=raw_source,
            user_id_column=0,
            item_id_column=1,
            score_column=4,
            score_processor=NumberNormalizer(scale=(1, 5))
        )

        score_result = rat.score_column

        self.assertTrue(-1 <= score <= 1 for score in score_result)

    # def test_add_score_column(self):
    #     ri = Ratings(
    #         source=raw_source,
    #         user_id_column=0,
    #         item_id_column=1,
    #         score_column=4
    #     )
    # 
    #     expected_columns = ['user_id', 'item_id', 'score']
    #     result_columns = ri.columns
    # 
    #     self.assertEqual(expected_columns, result_columns)
    # 
    #     ri.add_score_column('stars', column_name='score_duplicate')
    # 
    #     expected_columns = ['user_id', 'item_id', 'score', 'score_duplicate']
    #     result_columns = ri.columns
    # 
    #     self.assertEqual(expected_columns, result_columns)
    # 
    #     expected = [float(row['stars']) for row in raw_source_content]
    # 
    #     score_column_added_frame = list(ri.frame['score_duplicate'])
    #     score_column_added_dict = [rating_tuple[2] for user_value in ri.dict.values() for rating_tuple in user_value]
    # 
    #     self.assertEqual(expected, score_column_added_frame)
    #     self.assertEqual(expected, score_column_added_dict)
    # 
    # def test_add_score_column_w_score_processor(self):
    #     ri = Ratings(
    #         source=raw_source,
    #         user_id_column=0,
    #         item_id_column=1,
    #         score_column=4
    #     )
    # 
    #     ratings = ri.frame
    # 
    #     expected_columns = ['user_id', 'item_id', 'score']
    #     result_columns = ri.columns
    # 
    #     self.assertEqual(expected_columns, result_columns)
    # 
    #     ri.add_score_column('review_title', column_name='text_polarity',
    #                         score_processor=TextBlobSentimentAnalysis())
    # 
    #     expected_columns = ['user_id', 'item_id', 'score', 'text_polarity']
    #     result_columns = ri.columns
    # 
    #     self.assertEqual(expected_columns, result_columns)
    # 
    #     score_column_added = list(ratings['text_polarity'])
    #     score_column_added_dict = [rating_tuple[2] for user_value in ri.dict.values() for rating_tuple in user_value]
    # 
    #     self.assertTrue(-1 <= score <= 1 for score in score_column_added)
    #     self.assertTrue(-1 <= score <= 1 for score in score_column_added_dict)

    def test_ratings_to_csv(self):
        ri = Ratings(
            source=raw_source,
            user_id_column=0,
            item_id_column=1,
            score_column=4
        )

        # Test save
        ri.to_csv('csv_test/')
        self.assertTrue(os.path.isfile('csv_test/ratings_frame.csv'))

        # Test save first duplicate
        ri.to_csv('csv_test/')
        self.assertTrue(os.path.isfile('csv_test/ratings_frame (1).csv'))

        # Test save second duplicate
        ri.to_csv('csv_test/')
        self.assertTrue(os.path.isfile('csv_test/ratings_frame (2).csv'))

        # Test save with overwrite
        ri.to_csv('csv_test/', overwrite=True)
        self.assertTrue(os.path.isfile('csv_test/ratings_frame.csv'))
        self.assertFalse(os.path.isfile('csv_test/ratings_frame (3).csv'))

        # Test save with custom name
        ri.to_csv('csv_test/', 'ratings_custom_name')
        self.assertTrue(os.path.isfile('csv_test/ratings_custom_name.csv'))

        # remove test folder
        shutil.rmtree('csv_test/')

    def test_from_to_dataframe(self):
        df_ratings = pd.DataFrame({
            'user_id': ['u1', 'u1', 'u2', 'u3'],
            'item_id': ['i1', 'i2', 'i1', 'i80'],
            'score': [2, 3, 4, 1]
        })

        rat = Ratings.from_dataframe(df_ratings)

        self.assertEqual(rat.user_id_column, list(df_ratings['user_id']))
        self.assertEqual(rat.item_id_column, list(df_ratings['item_id']))
        self.assertEqual(rat.score_column, list(df_ratings['score']))

        # df_result = rat.to_dataframe()
        #
        # self.assertEqual(list(df_ratings['user_id']), list(df_result['user_id']))
        # self.assertEqual(list(df_ratings['item_id']), list(df_result['item_id']))
        # self.assertEqual(list(df_ratings['score']), list(df_result['score']))
        #
        # # raise exception when a column doesn't exist in the original df by key
        # with self.assertRaises(KeyError):
        #     Ratings.from_dataframe(df_ratings, user_column='not existent')
        #
        # # raise exception when a column doesn't exist in the original df by index
        # with self.assertRaises(IndexError):
        #     Ratings.from_dataframe(df_ratings, user_column=10)

    def test_from_list(self):
        list_no_timestamp = [Interaction('u1', 'i1', 5),
                             Interaction('u1', 'i2', 4),
                             Interaction('u2', 'i1', 2)]

        rat = Ratings.from_list(list_no_timestamp)

        self.assertEqual(rat.user_id_column, ['u1', 'u1', 'u2'])
        self.assertEqual(rat.item_id_column, ['i1', 'i2', 'i1'])
        self.assertEqual(rat.score_column, [5, 4, 2])
        self.assertTrue(len(rat.timestamp_column) == 0)

        list_w_timestamp = [Interaction('u1', 'i1', 5, 'timestamp1'),
                            Interaction('u1', 'i2', 4, 'timestamp2'),
                            Interaction('u2', 'i1', 2, 'timestamp3')]

        rat = Ratings.from_list(list_w_timestamp)

        self.assertEqual(rat.user_id_column, ['u1', 'u1', 'u2'])
        self.assertEqual(rat.item_id_column, ['i1', 'i2', 'i1'])
        self.assertEqual(rat.score_column, [5, 4, 2])
        self.assertEqual(rat.timestamp_column, ['timestamp1', 'timestamp2', 'timestamp3'])

    def test_from_dict(self):
        dict_no_timestamp = {'u1': [Interaction('u1', 'i1', 5),
                                    Interaction('u1', 'i2', 4)],
                             'u2': [Interaction('u2', 'i1', 2)]}

        rat = Ratings.from_dict(dict_no_timestamp)

        self.assertEqual(rat.user_id_column, ['u1', 'u1', 'u2'])
        self.assertEqual(rat.item_id_column, ['i1', 'i2', 'i1'])
        self.assertEqual(rat.score_column, [5, 4, 2])
        self.assertTrue(len(rat.timestamp_column) == 0)

        dict_w_timestamp = {'u1': [Interaction('u1', 'i1', 5, 'timestamp1'),
                                   Interaction('u1', 'i2', 4, 'timestamp2')],
                            'u2': [Interaction('u2', 'i1', 2, 'timestamp3')]}

        rat = Ratings.from_dict(dict_w_timestamp)

        self.assertEqual(rat.user_id_column, ['u1', 'u1', 'u2'])
        self.assertEqual(rat.item_id_column, ['i1', 'i2', 'i1'])
        self.assertEqual(rat.score_column, [5, 4, 2])
        self.assertEqual(rat.timestamp_column, ['timestamp1', 'timestamp2', 'timestamp3'])

    def test_get_user_interaction(self):
        df_ratings = pd.DataFrame({
            'user_id': ['u1', 'u1', 'u1', 'u2'],
            'item_id': ['i1', 'i2', 'i3', 'i80'],
            'score': [2, 3, 4, 1]
        })

        rat = Ratings.from_dataframe(df_ratings)

        user_interactions = rat.get_user_interactions('u1')

        self.assertEqual(['i1', 'i2', 'i3'], [interaction.item_id for interaction in user_interactions])
        self.assertEqual([2, 3, 4], [interaction.score for interaction in user_interactions])

        # get first 2 user interactions
        user_interactions = rat.get_user_interactions('u1', head=2)

        self.assertEqual(['i1', 'i2'], [interaction.item_id for interaction in user_interactions])
        self.assertEqual([2, 3], [interaction.score for interaction in user_interactions])

    def test_filter_ratings(self):
        df_ratings = pd.DataFrame({
            'user_id': ['u1', 'u1', 'u2', 'u3'],
            'item_id': ['i1', 'i2', 'i3', 'i80'],
            'score': [2, 3, 4, 1]
        })

        rat = Ratings.from_dataframe(df_ratings)

        rat_filtered = rat.filter_ratings(['u1', 'u3'])

        self.assertEqual(['u1', 'u1', 'u3'], rat_filtered.user_id_column)
        self.assertEqual(['i1', 'i2', 'i80'], rat_filtered.item_id_column)
        self.assertEqual([2, 3, 1], rat_filtered.score_column)

    def test_take_head_all(self):
        df_ratings = pd.DataFrame({
            'user_id': ['u1', 'u1', 'u2', 'u2'],
            'item_id': ['i1', 'i2', 'i3', 'i80'],
            'score': [2, 3, 4, 1]
        })

        rat = Ratings.from_dataframe(df_ratings)

        # take first element for all users
        head_rat = rat.take_head_all(1)

        self.assertEqual(['u1', 'u2'], head_rat.user_id_column)
        self.assertEqual(['i1', 'i3'], head_rat.item_id_column)
        self.assertEqual([2, 4], head_rat.score_column)

    def test_exception_import_ratings(self):
        # Test exception column name not present in raw source
        with self.assertRaises(KeyError):
            Ratings(
                source=raw_source,
                user_id_column='not_existent',
                item_id_column='item_id',
                score_column='stars')

        # Test exception column index not present in raw source
        with self.assertRaises(IndexError):
            Ratings(
                source=raw_source,
                user_id_column=99,
                item_id_column='item_id',
                score_column='stars')

        # Test exception score column can't be converted into float
        with self.assertRaises(ValueError):
            Ratings(
                source=raw_source,
                user_id_column='user_id',
                item_id_column='item_id',
                score_column='review_title')

    # def test_exception_add_score_column(self):
    #     # Test exception score column can't be converted into float
    #     ri = Ratings(
    #         source=raw_source,
    #         user_id_column='user_id',
    #         item_id_column='item_id',
    #         score_column='stars')
    #
    #     with self.assertRaises(ValueError):
    #         ri.add_score_column('review_title', 'text')


if __name__ == '__main__':
    unittest.main()



import unittest
from unittest import TestCase
import pandas as pd

from clayrs.content_analyzer.ratings_manager.ratings import Ratings
from clayrs.recsys.partitioning import HoldOutPartitioning, KFoldPartitioning, BootstrapPartitioning

original_ratings = pd.DataFrame.from_dict(
    {'from_id': ["001", "001", "002", "002", "002", "003", "003", "004", "004"],
     'to_id': ["aaa", "bbb", "aaa", "ddd", "ccc", "ccc", "aaa", "ddd", "ccc"],
     'rating': [0.8, 0.7, -0.4, 1.0, 0.4, 0.1, -0.3, 0.5, 0.7]})

original_ratings = Ratings.from_dataframe(original_ratings)


class TestPartitioning(TestCase):
    def check_partition_correct(self, train, test, original):
        original_list = list(original)
        train_list = list(train)
        test_list = list(test)

        # Check that train and test are a partition
        train_not_in_test = [row for row in train_list if row not in test_list]
        self.assertCountEqual(train_list, train_not_in_test)  # Count so regardless of order
        test_not_in_train = [row for row in test_list if row not in train_list]
        self.assertCountEqual(test_list, test_not_in_train)  # Count so regardless of order

        # Check that the union of the two give the original data
        union_list = train_list + test_list
        self.assertCountEqual(original_list, union_list)  # Count so regardless of order


class TestKFoldPartitioning(TestPartitioning):

    def test_split_all(self):

        kf = KFoldPartitioning(n_splits=2)

        result_train, result_test = kf.split_all(original_ratings)

        for train, test in zip(result_train, result_test):
            self.check_partition_correct(train, test, original_ratings)

        result_train, result_test = kf.split_all(original_ratings, user_id_list={'001'})

        for train, test in zip(result_train, result_test):
            only_001_train = set(train.user_id_column)
            only_001_test = list(test.user_id_column)

            self.assertTrue(len(only_001_train) == 1)
            self.assertTrue(len(only_001_test) == 1)

            self.assertIn("001", only_001_train)
            self.assertIn("001", only_001_test)

            original_001 = original_ratings.filter_ratings(user_list=['001'])

            self.check_partition_correct(train, test, original_001)

    def test_split_some_missing(self):
        kf = KFoldPartitioning(n_splits=3)

        result_train, result_test = kf.split_all(original_ratings)

        for train, test in zip(result_train, result_test):
            # It's the only user for which is possible to perform 3 split
            only_002_train = set(train.user_id_column)
            only_002_test = set(test.user_id_column)

            self.assertTrue(len(only_002_train) == 1)
            self.assertTrue(len(only_002_test) == 1)

            self.assertIn("002", only_002_train)
            self.assertIn("002", only_002_test)

            original_002 = original_ratings.filter_ratings(user_list=['002'])

            self.check_partition_correct(train, test, original_002)

    def test_split_all_raise_error(self):
        kf = KFoldPartitioning(n_splits=3, skip_user_error=False)

        with self.assertRaises(ValueError):
            kf.split_all(original_ratings)


class TestHoldOutPartitioning(TestPartitioning):

    def test_split_all(self):
        hold_percentage = 0.3

        ho = HoldOutPartitioning(train_set_size=hold_percentage)

        result_train, result_test = ho.split_all(original_ratings)

        for train, test in zip(result_train, result_test):
            self.check_partition_correct(train, test, original_ratings)

            train_percentage = (len(train) / len(original_ratings))
            self.assertEqual(hold_percentage, train_percentage)

        result_train, result_test = ho.split_all(original_ratings, user_id_list={'001'})

        for train, test in zip(result_train, result_test):
            only_001_train = set(train.user_id_column)
            only_001_test = set(test.user_id_column)

            self.assertTrue(len(only_001_train) == 1)
            self.assertTrue(len(only_001_test) == 1)

            self.assertIn("001", only_001_train)
            self.assertIn("001", only_001_test)

            original_001 = original_ratings.filter_ratings(user_list=['001'])

            self.check_partition_correct(train, test, original_001)

    def test_split_some_missing(self):

        hold_percentage = 0.4

        ho = HoldOutPartitioning(train_set_size=hold_percentage)

        result_train, result_test = ho.split_all(original_ratings)

        for train, test in zip(result_train, result_test):
            # It's the only user for which is possible to hold 0.4 as train
            only_002_train = set(train.user_id_column)
            only_002_test = set(test.user_id_column)

            self.assertTrue(len(only_002_train) == 1)
            self.assertTrue(len(only_002_test) == 1)

            self.assertIn("002", only_002_train)
            self.assertIn("002", only_002_test)

            original_002 = original_ratings.filter_ratings(user_list=['002'])

            self.check_partition_correct(train, test, original_002)

    def test_split_all_raise_error(self):
        hold_percentage = 0.4

        ho = HoldOutPartitioning(train_set_size=hold_percentage, skip_user_error=False)

        with self.assertRaises(ValueError):
            ho.split_all(original_ratings)

    def test__check_percentage(self):

        # Bigger than 1
        hold_percentage = 1.5
        with self.assertRaises(ValueError):
            HoldOutPartitioning(train_set_size=hold_percentage)

        # Negative percentage
        hold_percentage = -0.2
        with self.assertRaises(ValueError):
            HoldOutPartitioning(train_set_size=hold_percentage)


class TestBootstrapPartitioning(TestPartitioning):

    def check_partition_correct(self, train, test, original):
        original_list = list(original)
        train_list = list(train)
        test_list = list(test)

        # Check that train and test are a partition
        train_not_in_test = [row for row in train_list if row not in test_list]
        self.assertCountEqual(train_list, train_not_in_test)  # Count so regardless of order
        test_not_in_train = [row for row in test_list if row not in train_list]
        self.assertCountEqual(test_list, test_not_in_train)  # Count so regardless of order

        # Check that the union of the two give the original data.
        # We remove any duplicate that can naturally happen due to the resampling of the
        # bootstrap method
        union_list = list(set(train_list)) + test_list
        self.assertCountEqual(original_list, union_list)  # Count so regardless of order

    def test_split_all(self):

        bs = BootstrapPartitioning(random_state=5)

        [train], [test] = bs.split_all(original_ratings)

        # 001, 003 and 004 had not enough ratings and so with this particular random state
        # the resampling will give us empty test set for those user, meaning that they will not be
        # present in the final train and test set
        self.assertTrue("001" not in train.user_id_column)
        self.assertTrue("002" in train.user_id_column)
        self.assertTrue("003" not in train.user_id_column)
        self.assertTrue("004" not in train.user_id_column)

        # only user 002 is present in train and test set
        original_002 = original_ratings.filter_ratings(['002'])
        self.check_partition_correct(train, test, original_002)

    def test_split_raise_error(self):
        bs = BootstrapPartitioning(random_state=5)

        # with this particular random state user 001 has not enough ratings and the resampling
        # will get all its ratings, making the test set empty, and so an error will be raised
        # by the split_single method
        user_001_rat = original_ratings.get_user_interactions('001')

        with self.assertRaises(ValueError):
            bs.split_single(user_001_rat)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from typing import List, Iterable, TYPE_CHECKING

if TYPE_CHECKING:
    from clayrs.recsys.graphs.feature_selection.feature_selection_alg import FeatureSelectionAlgorithm
    from clayrs.recsys.graphs.graph import FullDiGraph, UserNode, ItemNode

from clayrs.recsys.graphs.feature_selection.exceptions import FeatureSelectionException
from clayrs.utils.const import logger


def feature_selector(graph: FullDiGraph,
                     fs_algorithm_user: FeatureSelectionAlgorithm = None,
                     fs_algorithm_item: FeatureSelectionAlgorithm = None,
                     user_target_nodes: Iterable[UserNode] = None,
                     item_target_nodes: Iterable[ItemNode] = None,
                     inplace: bool = False) -> FullDiGraph:
    """
    Given a FullGraph, this method performs feature selection on it and returns the "reduced" graph.

    You can choose to reduce only *user properties* (*evaluate the `fs_algorithm_user` parameter*),
    to reduce only *item properties* (*evaluate the `fs_algorithm_item` parameter*) or both (*evaluate
    the `fs_algorithm_user` parameter* and the `fs_algorithm_item` parameter*).
    You can also choose different *feature selection algorithms* for users and items.

    You can also define a custom list of user and item nodes:

    * In this case only properties of those nodes will be considered during the feature selection process (instead of
    using properties of all users and items)

    This function changes a *copy* of the original graph by default, but you can change this behaviour with the
    `inplace` parameter.

    Examples:

        ```python
        # create a full graph
        full_graph = rs.NXFullGraph(ratings,
                                     user_contents_dir='users_codified/', # (1)
                                     item_contents_dir='movies_codified/', # (2)
                                     user_exo_properties={0}, # (3)
                                     item_exo_properties={'dbpedia'}, # (4)
                                     link_label='score')

         # perform feature selection by keeping only top 5 property labels
         # according to page rank algorithm
         fs_graph = rs.feature_selector(full_graph,
                                        fs_algorithm_item=rs.TopKPageRank(k=5))
        ```

    Args:
        graph: Original graph on which feature selection will be performed
        fs_algorithm_user: FeatureSelectionAlgorithm that will be performed on user properties. Can be different from
            `fs_algorithm_item`
        fs_algorithm_item: FeatureSelectionAlgorithm that will be performed on item properties. Can be different from
            `fs_algorithm_user`
        user_target_nodes (list): List of user nodes to consider in the feature selection process: only properties
            of user nodes in this list will be "reduced"
        item_target_nodes (list): List of item nodes to consider in the feature selection process: only properties
            of item nodes in this list will be "reduced"
        inplace: Boolean parameter that let you choose if changes must be performed on the original graph
            (`inplace=True`) or on its copy (`inplace=False`). Default is False

    Returns:
        Copy of the original graph from which the less important Property nodes (the ones having edges with less
        important property labels) will be removed
    """
    if fs_algorithm_user is not None and user_target_nodes is None:
        user_target_nodes = graph.user_nodes

    if fs_algorithm_item is not None and item_target_nodes is None:
        item_target_nodes = graph.item_nodes

    property_labels_to_remove = list()
    user_fs_failed = False
    item_fs_failed = False

    if fs_algorithm_user is not None:
        logger.info("Performing Feature Selection on users")
        try:
            user_props_to_remove = fs_algorithm_user.perform(graph, list(user_target_nodes), mode='to_remove')
            property_labels_to_remove.extend(user_props_to_remove)
        except FeatureSelectionException as e:
            logger.warning(str(e) + "!\nUsers original properties will be kept")
            user_fs_failed = True

    if fs_algorithm_item is not None:
        logger.info("Performing Feature Selection on items")
        try:
            item_props_to_remove = fs_algorithm_item.perform(graph, list(item_target_nodes), mode='to_remove')
            property_labels_to_remove.extend(item_props_to_remove)
        except FeatureSelectionException as e:
            logger.warning(str(e) + "!\nItems original properties will be kept")
            item_fs_failed = True

    # in case user feature selection or item feature selection failed
    # if both failed the original graph is returned
    # if only one of them failed, the original properties (either for items or users) are retrieved
    if user_fs_failed and item_fs_failed:
        logger.warning("Since both feature selection on items and feature selection on users failed or no fs algorithm"
                       "has been defined,\nthe original graph will be returned")

    if inplace is True:
        graph_fs = _delete_property_nodes(graph, property_labels_to_remove)
    else:
        graph_copy = graph.copy()
        graph_fs = _delete_property_nodes(graph_copy, property_labels_to_remove)

    return graph_fs


def _delete_property_nodes(graph: FullDiGraph, property_labels_to_remove: List[str]) -> FullDiGraph:
    """
    Creates a copy of the original graph from which the Property nodes with labels defined in the
    `properties_to_remove` parameter will be deleted (these property nodes will be the ones for which the feature
    selection algorithm found the lowest 'importance' score for the property label of their links)

    Args:
        graph: The original graph used for Feature Selection
        property_labels_to_remove: List of property labels that should be removed from the original graph

    Returns:
        Graph on which the less important property nodes will be removed
    """
    # MAYBE delete only edges and then if a node is "alone" delete it? In this way we preserve the cases in which
    # the PropertyNode DiCaprio can be linked with an edge "director" and an edge "actor" and we need to remove
    # "director" properties

    prop_nodes_to_remove = []
    for property_node in graph.property_nodes:
        predecessors = graph.get_predecessors(property_node)
        for pred in predecessors:
            link_label = graph.get_link_data(pred, property_node).get('label')
            if link_label in set(property_labels_to_remove):
                prop_nodes_to_remove.append(property_node)

                # we don't check for other predecessors,
                # if node is linked to only one node with the label to remove it will be removed
                break

    graph.remove_node(prop_nodes_to_remove)
    return graph



from __future__ import annotations
import itertools
from typing import List, Set, Any, Union, TYPE_CHECKING

import networkx as nx

if TYPE_CHECKING:
    from clayrs.content_analyzer.ratings_manager.ratings import Ratings
    from clayrs.recsys.graphs import NXBipartiteGraph
    from clayrs.recsys.methodology import Methodology

from clayrs.content_analyzer.ratings_manager.ratings import Interaction
from clayrs.recsys.graph_based_algorithm.page_rank.page_rank import PageRank
from clayrs.recsys.graphs.graph import UserNode, ItemNode
from clayrs.recsys.methodology import TestRatingsMethodology
from clayrs.utils.context_managers import get_iterator_parallel


class NXPageRank(PageRank):
    """
    Page Rank algorithm based on the networkx implementation.
    Please note that it can only be used for instantiated NXGraphs

    The PageRank can be ***personalized***, in this case the PageRank will be calculated with a personalization vector
    made by items in the user profile weighted by the score given to them.

    Args:
        alpha: Damping parameter for PageRank, default=0.85.
        personalized: Boolean value that specifies if the page rank must be calculated considering the user profile
            as personalization vector. Default is False
        max_iter: Maximum number of iterations in power method eigenvalue solver.
        tol: Error tolerance used to check convergence in power method solver.
        nstart: Starting value of PageRank iteration for each node.
        weight: Boolean value which tells the algorithm if weight of the edges must be considered or not.
            Default is True
    """

    def __init__(self, alpha: Any = 0.85, personalized: bool = False, max_iter: Any = 100, tol: Any = 1.0e-6,
                 nstart: Any = None, weight: bool = True):
        self.alpha = alpha
        self.max_iter = max_iter
        self.tol = tol
        self.nstart = nstart
        self.weight = weight

        super().__init__(personalized)

    def rank(self, all_users: Set[str], graph: NXBipartiteGraph, test_set: Ratings,
             recs_number: int = None, methodology: Union[Methodology, None] = TestRatingsMethodology(),
             num_cpus: int = 1) -> List[Interaction]:
        """
        Rank the top-n recommended items for the user. If the recs_number parameter isn't specified,
        All unrated items for the user will be ranked (or only items in the filter list, if specified).

        One can specify which items must be ranked for each user with the `filter_dict` parameter,
        in this case every user is mapped with a list of items for which a ranking score must be computed.
        Otherwise, **ALL** unrated items will be ranked for each user.

        Args:
            all_users: Set of user id for which a recommendation list must be generated
            graph: A NX graph previously instantiated
            recs_number: number of the top ranked items to return, if None all ranked items will be returned
            test_set: Ratings object which represents the ground truth of the split considered
            recs_number: number of the top ranked items to return, if None all ranked items will be returned
            methodology: `Methodology` object which governs the candidate item selection. Default is
                `TestRatingsMethodology`
            num_cpus: number of processors that must be reserved for the method. Default is 0, meaning that
                the number of cpus will be automatically detected.

        Returns:
            List of Interactions object in a descending order w.r.t the 'score' attribute, representing the ranking for
                a single user
        """

        def compute_single_rank(user_id):

            # nonlocal keyword allows to modify the score variable
            nonlocal scores

            user_node = UserNode(user_id)

            filter_list = None
            if methodology is not None:
                filter_list = set(ItemNode(item_to_rank) for item_to_rank in
                                  methodology.filter_single(user_id, train_set, test_set))

            # run the pageRank
            if self._personalized is True:
                # the personalization vector is formed by the nodes that the user voted with their weight
                # + all the other nodes in the graph with weight as the min weight given by the user
                # (This because if a node isn't specified in the personalization vector will have 0 score in page
                # rank)
                succ = graph.get_successors(user_node)
                profile = {scored_node: graph.get_link_data(user_node, scored_node).get('weight')
                           for scored_node in succ
                           if graph.get_link_data(user_node, scored_node).get('weight') is not None}

                pers = {node: profile[node] if node in profile else min(set(profile.values()))
                        for node in graph.to_networkx().nodes}

                scores = nx.pagerank(graph.to_networkx(), personalization=pers, alpha=self.alpha,
                                     max_iter=self.max_iter, tol=self.tol, nstart=self.nstart, weight=weight)

            # if scores is None it means this is the first time we are running normal pagerank
            # for all the other users the pagerank won't be computed again
            elif scores is None:
                scores = nx.pagerank(graph.to_networkx(), alpha=self.alpha, max_iter=self.max_iter,
                                     tol=self.tol, nstart=self.nstart, weight=weight)

            # clean the results removing user nodes, selected user profile and eventually properties
            user_scores = self.filter_result(graph, scores, filter_list, user_node)

            # Build the item_score dict (key is item_id, value is rank score predicted)
            # and order the keys in descending order
            item_score_dict = dict(zip([node.value for node in user_scores.keys()], user_scores.values()))
            ordered_item_ids = sorted(item_score_dict, key=item_score_dict.get, reverse=True)

            # we only save the top-n items_ids corresponding to top-n recommendations
            # (if recs_number is None ordered_item_ids will contain all item_ids as the original list)
            ordered_item_ids = ordered_item_ids[:recs_number]

            # we construct the output data
            single_rank_interaction_list = [Interaction(user_id, item_id, item_score_dict[item_id])
                                            for item_id in ordered_item_ids]

            return user_id, single_rank_interaction_list

        # scores will contain pagerank scores
        scores = None
        all_rank_interaction_list = []
        weight = 'weight' if self.weight is True else None
        train_set = graph.to_ratings()

        with get_iterator_parallel(num_cpus,
                                   compute_single_rank, all_users,
                                   progress_bar=True, total=len(all_users)) as pbar:

            pbar.set_description("Prepping rank...")

            for user_id, user_rank in pbar:
                all_rank_interaction_list.append(user_rank)
                pbar.set_description(f"Computing rank for user {user_id}")

        return list(itertools.chain.from_iterable(all_rank_interaction_list))

    def __str__(self):
        return "NXPageRank"

    def __repr__(self):
        return f"NXPageRank(alpha={self.alpha}, personalized={self._personalized}, max_iter={self.max_iter}, " \
               f"tol={self.tol}, nstart={self.nstart}, weight={self.weight})"



from gensim import downloader
import numpy as np

from clayrs.utils.const import logger
from clayrs.content_analyzer.embeddings.embedding_loader.embedding_loader import WordEmbeddingLoader


class Gensim(WordEmbeddingLoader):
    """
    Class that produces word embeddings using gensim pre-trained models.

    The model will be automatically downloaded using the gensim downloader api if not present locally.

    Args:
        model_name: Name of the model to load/download
    """

    def __init__(self, model_name: str = 'glove-twitter-25'):
        super().__init__(model_name)

    def get_vector_size(self) -> int:
        return self.model.vector_size

    def get_embedding(self, word: str) -> np.ndarray:
        return self.model[word]

    def load_model(self):
        # if the reference isn't in the possible models, FileNotFoundError is raised
        if self.reference in downloader.info()['models']:
            logger.info(f"Downloading/Loading {str(self)}")

            return downloader.load(self.reference)
        else:
            raise FileNotFoundError

    def __str__(self):
        return f"Gensim {self.reference}"

    def __repr__(self):
        return f'Gensim(model_name={self.reference}'



import pandas as pd
from unittest import TestCase
import os

from clayrs.content_analyzer import Ratings
from clayrs.recsys import PropertyNode
from clayrs.recsys.graphs.feature_selection import TopKPageRank, TopKDegreeCentrality, \
    TopKEigenVectorCentrality
from clayrs.recsys.graphs.feature_selection import FeatureSelectionException
from clayrs.recsys.graphs import NXFullGraph
from test import dir_test_files


class TestFeatureSelectionAlgorithm(TestCase):

    def setUp(self) -> None:
        movies_dir = os.path.join(dir_test_files, 'complex_contents', 'movies_codified/')
        user_dir = os.path.join(dir_test_files, 'complex_contents', 'users_codified/')

        df = pd.DataFrame.from_dict({'user_id': ["1", "1", "2", "2", "2", "3", "4", "4"],
                                     'item_id': ["tt0112281", "tt0113041", "tt0112453", "tt0112346",
                                                 "tt0112302", "tt0113101", "tt0113277", "tt0113228"],
                                     'score': [5, 3, 1, 5, 4, 2, 2, 4]})
        rat = Ratings.from_dataframe(df)

        # all properties from the dbpedia item repr extracted
        self.g: NXFullGraph = NXFullGraph(rat,
                                          user_contents_dir=user_dir,
                                          item_contents_dir=movies_dir,
                                          item_exo_properties={'dbpedia'},
                                          user_exo_properties={'local': '1'}
                                          )
        # we use as target nodes all item nodes in the graph, meaning that we want want to remove
        # less important features of all the item nodes of the graph
        self.target_item_nodes = list(self.g.item_nodes)

        # here we save in a set all unique property labels such as 'starring', 'editing', etc.
        self.unique_prop_labels = set()
        for item_node in self.target_item_nodes:
            prop_nodes = [node for node in self.g.get_successors(item_node) if isinstance(node, PropertyNode)]
            for prop_n in prop_nodes:
                link_label = self.g.get_link_data(item_node, prop_n)['label']
                self.unique_prop_labels.add(link_label)

        # 'film director' is one of the label that appears less often, so, since the example graph is pretty simple,
        # the result of feature selection algorithms should not include it when the mode is 'to_remove'
        self.less_important_property_label = 'film director'

        # 'starring' is one of the label that appears more often, so, since the example graph is pretty simple,
        # the result of feature selection algorithms should include it when the mode is 'to_keep'
        self.most_important_property_label = 'starring'

    def test_perform_NXTopKPageRank(self):
        top_prop_labels_to_keep = 3

        result = TopKPageRank(k=top_prop_labels_to_keep).perform(self.g, self.target_item_nodes, mode='to_remove')

        # mode -> 'to_remove' will return a list of the property labels to remove
        n_nodes_to_remove = len(self.unique_prop_labels) - top_prop_labels_to_keep
        self.assertEqual(len(result), n_nodes_to_remove)
        self.assertTrue(self.less_important_property_label in result)
        self.assertTrue(self.most_important_property_label not in result)

        result = TopKPageRank(k=top_prop_labels_to_keep).perform(self.g, self.target_item_nodes, mode='to_keep')

        # mode -> 'to_keep' will return a list of the property labels to keep
        self.assertEqual(len(result), top_prop_labels_to_keep)
        self.assertTrue(self.less_important_property_label not in result)
        self.assertTrue(self.most_important_property_label in result)

    def test_perform_NXTopKDegreeCentrality(self):
        top_prop_labels_to_keep = 3
        result = TopKDegreeCentrality(k=top_prop_labels_to_keep).perform(self.g, self.target_item_nodes,
                                                                         mode='to_remove')

        # mode -> 'to_remove' will return a list of the property labels to remove
        n_nodes_to_remove = len(self.unique_prop_labels) - top_prop_labels_to_keep
        self.assertEqual(len(result), n_nodes_to_remove)
        self.assertTrue(self.less_important_property_label in result)
        self.assertTrue(self.most_important_property_label not in result)

        result = TopKDegreeCentrality(k=top_prop_labels_to_keep).perform(self.g, self.target_item_nodes,
                                                                         mode='to_keep')

        # mode -> 'to_keep' will return a list of the property labels to keep
        self.assertEqual(len(result), top_prop_labels_to_keep)
        self.assertTrue(self.less_important_property_label not in result)
        self.assertTrue(self.most_important_property_label in result)

    def test_perform_NXTopKEigenVectorCentrality(self):
        top_prop_labels_to_keep = 3
        result = TopKEigenVectorCentrality(k=top_prop_labels_to_keep, max_iter=200).perform(self.g,
                                                                                            self.target_item_nodes,
                                                                                            mode='to_remove')

        # mode -> 'to_remove' will return a list of the property labels to remove
        n_nodes_to_remove = len(self.unique_prop_labels) - top_prop_labels_to_keep
        self.assertEqual(len(result), n_nodes_to_remove)
        self.assertTrue(self.less_important_property_label in result)
        self.assertTrue(self.most_important_property_label not in result)

        result = TopKEigenVectorCentrality(k=top_prop_labels_to_keep, max_iter=200).perform(self.g,
                                                                                            self.target_item_nodes,
                                                                                            mode='to_keep')

        # mode -> 'to_keep' will return a list of the property labels to keep
        self.assertEqual(len(result), top_prop_labels_to_keep)
        self.assertTrue(self.less_important_property_label not in result)
        self.assertTrue(self.most_important_property_label in result)

    def test_special_cases_feature_selection(self):
        # empty target nodes
        self.assertEqual(TopKPageRank(k=2).perform(self.g, []), [])

        # k = 0
        result = TopKPageRank(k=0).perform(self.g, self.target_item_nodes)
        self.assertEqual(result, [])

        # k < 0
        with self.assertRaises(ValueError):
            TopKPageRank(k=-1)

        # k > number of properties in the graph
        result = TopKPageRank(100).perform(self.g, self.target_item_nodes)
        self.assertEqual(len(result), len(self.unique_prop_labels))  # number of item labels

        # method fails to converge
        with self.assertRaises(FeatureSelectionException):
            TopKPageRank(k=1, max_iter=0).perform(self.g, self.target_item_nodes)

        # non existent mode
        with self.assertRaises(TypeError):
            TopKPageRank().perform(self.g, [], mode='invalid')



import os

from whoosh.analysis import SimpleAnalyzer
from whoosh.fields import Schema, TEXT, KEYWORD
from whoosh.index import create_in, open_dir
from whoosh.formats import Frequency
from whoosh.qparser import QueryParser, OrGroup, FieldsPlugin
from whoosh.query import Term, Or
from whoosh.scoring import TF_IDF, BM25F
from typing import Union, Dict
import math
import abc

from clayrs.content_analyzer.memory_interfaces.memory_interfaces import TextInterface


class IndexInterface(TextInterface):
    """
    Abstract class that takes care of serializing and deserializing text in an indexed structure
    using the Whoosh library

    Args:
        directory (str): Path of the directory where the content will be serialized
    """

    def __init__(self, directory: str):
        super().__init__(directory)
        self.__doc = None  # document that is currently being created and will be added to the index
        self.__writer = None  # index writer
        self.__doc_index = 0  # current position the document will have in the index once it is serialized
        self.__schema_changed = False  # true if the schema has been changed, false otherwise

    @property
    @abc.abstractmethod
    def schema_type(self):
        """
        Whoosh uses a Schema that defines, for each field of the content, how to store the data. In the case of this
        project, every field will have the same structure and will share the same field type. This method returns
        said field type.
        """
        raise NotImplementedError

    def init_writing(self, delete_old: bool = False):
        """
        Creates the index locally (in the directory passed in the constructor) and initializes the index writer.
        If an index already exists in the directory, what happens depend on the attribute delete_old passed as argument

        Args:
            delete_old (bool): if True, the index that was in the same directory is destroyed and replaced;
                if False, the index is simply opened
        """
        if os.path.exists(self.directory):
            if delete_old:
                self.delete()
                os.mkdir(self.directory)
                ix = create_in(self.directory, Schema())
                self.__writer = ix.writer()
            else:
                ix = open_dir(self.directory)
                self.__writer = ix.writer()
                self.__doc_index = self.__writer.reader().doc_count()
        else:
            os.mkdir(self.directory)
            ix = create_in(self.directory, Schema())
            self.__writer = ix.writer()

    def new_content(self):
        """
        The new content is a document that will be indexed. In this case the document is a dictionary with
        the name of the field as key and the data inside the field as value
        """
        self.__doc = {}

    def new_field(self, field_name: str, field_data: object):
        """
        Adds a new field to the document that is being created. Since the index Schema is generated dynamically, if
        the field name is not in the Schema already it is added to it

        Args:
            field_name (str): Name of the new field
            field_data (object): Data to put into the field
        """
        if field_name not in open_dir(self.directory).schema.names():
            self.__writer.add_field(field_name, self.schema_type)
            self.__schema_changed = True
        self.__doc[field_name] = field_data

    def serialize_content(self) -> int:
        """
        Serializes the content in the index. If the schema changed, the writer will commit the changes to the schema
        before adding the document to the index. Once the document is indexed, it can be deleted from the IndexInterface
        and the document position in the index is returned
        """
        if self.__schema_changed:
            self.__writer.commit()
            self.__writer = open_dir(self.directory).writer()
            self.__schema_changed = False
        self.__writer.add_document(**self.__doc)
        del self.__doc
        self.__doc_index += 1
        return self.__doc_index - 1

    def stop_writing(self):
        """
        Stops the index writer and commits the operations
        """
        self.__writer.commit()
        del self.__writer

    def get_field(self, field_name: str, content_id: Union[str, int]) -> str:
        """
        Uses a search index to retrieve the content corresponding to the content_id (if it is a string) or in the
        corresponding position (if it is an integer), and returns the data in the field corresponding to the field_name

        Args:
            field_name (str): name of the field from which the data will be retrieved
            content_id (Union[str, int]): either the position or Id of the content that contains the specified field

        Returns:
            Data contained in the field of the content
        """
        ix = open_dir(self.directory)
        with ix.searcher() as searcher:
            if isinstance(content_id, str):
                query = Term("content_id", content_id)
                result = searcher.search(query)
                result = result[0][field_name]
            elif isinstance(content_id, int):
                result = searcher.reader().stored_fields(content_id)[field_name]
            return result

    def query(self, string_query: str, results_number: int, mask_list: list = None,
              candidate_list: list = None, classic_similarity: bool = True) -> dict:
        """
        Uses a search index to query the index in order to retrieve specific contents using a query expressed in string
        form

        Args:
            string_query: query expressed as a string
            results_number: number of results the searcher will return for the query
            mask_list: list of content_ids of items to ignore in the search process
            candidate_list: list of content_ids of items to consider in the search process,
                if it is not None only items in the list will be considered
            classic_similarity: if True, classic tf idf is used for scoring, otherwise BM25F is used

        Returns:
            results: the final results dictionary containing the results found from the search index for the
                query. The dictionary will be in the following form:

                    {content_id: {"item": item_dictionary, "score": item_score}, ...}

                content_id is the content_id for the corresponding item
                item_dictionary is the dictionary of the item containing the fields as keys and the contents as values.
                So it will be in the following form:

                    {"Plot": "this is the plot", "Genre": "this is the Genre"}

                The item_dictionary will not contain the content_id since it is already defined and used as key of the
                external dictionary
                items_score is the score given to the item for the query by the index searcher
        """
        ix = open_dir(self.directory)
        with ix.searcher(weighting=TF_IDF if classic_similarity else BM25F) as searcher:
            candidate_query_list = None
            mask_query_list = None

            # the mask list contains the content_id for the items to ignore in the searching process
            # from the mask list a mask query is created and it will be used by the searcher
            if mask_list is not None:
                mask_query_list = []
                for document in mask_list:
                    mask_query_list.append(Term("content_id", document))
                mask_query_list = Or(mask_query_list)

            # the candidate list contains the content_id for the items to consider in the searching process
            # from the candidate list a candidate query is created and it will be used by the searcher
            if candidate_list is not None:
                candidate_query_list = []
                for candidate in candidate_list:
                    candidate_query_list.append(Term("content_id", candidate))
                candidate_query_list = Or(candidate_query_list)

            schema = ix.schema
            parser = QueryParser("content_id", schema=schema, group=OrGroup)
            # regular expression to match the possible field styles
            # examples: "content_id" or "Genre#2" or "Genre#2#custom_id"
            parser.add_plugin(FieldsPlugin(r'(?P<text>[\w-]+(\#[\w-]+(\#[\w-]+)?)?|[*]):'))
            query = parser.parse(string_query)
            score_docs = \
                searcher.search(query, limit=results_number, filter=candidate_query_list, mask=mask_query_list)

            # creation of the results dictionary, This phase is necessary because the Hit objects returned by the
            # searcher as results need the reader inside the search index in order to return information
            # so it would be impossible to access a field or the score of the item from outside this method
            # because of that this dictionary containing the most important infos is created
            results = {}
            for hit in score_docs:
                hit_dict = dict(hit)
                content_id = hit_dict.pop("content_id")
                results[content_id] = {}
                results[content_id]["item"] = hit_dict
                results[content_id]["score"] = hit.score
            return results

    def get_tf_idf(self, field_name: str, content_id: Union[str, int]) -> Dict[str, float]:
        r"""
        Calculates the tf-idf for the words contained in the field of the content whose id
        is content_id (if it is a string) or in the given position (if it is an integer).

        The tf-idf computation formula is:

        $$
        tf \mbox{-} idf = (1 + log10(tf)) * log10(idf)
        $$

        Args:
            field_name: Name of the field containing the words for which calculate the tf-idf
            content_id: either the position or Id of the content that contains the specified field

        Returns:
            words_bag: Dictionary whose keys are the words contained in the field, and the
                corresponding values are the tf-idf values
        """
        ix = open_dir(self.directory)
        words_bag = {}
        with ix.searcher() as searcher:
            if isinstance(content_id, str):
                query = Term("content_id", content_id)
                doc_num = searcher.search(query).docnum(0)
            elif isinstance(content_id, int):
                doc_num = content_id

            # if the document has the field == "" (length == 0) then the bag of word is empty
            if len(searcher.ixreader.stored_fields(doc_num)[field_name]) > 0:
                # retrieves the frequency vector (used for tf)
                list_with_freq = [term_with_freq for term_with_freq
                                  in searcher.vector(doc_num, field_name).items_as("frequency")]
                for term, freq in list_with_freq:
                    tf = 1 + math.log10(freq)
                    idf = math.log10(searcher.doc_count()/searcher.doc_frequency(field_name, term))
                    words_bag[term] = tf*idf
        return words_bag

    @abc.abstractmethod
    def __str__(self):
        raise NotImplementedError

    @abc.abstractmethod
    def __repr__(self):
        raise NotImplementedError


class KeywordIndex(IndexInterface):
    """
    This class implements the schema_type method: KeyWord. This is useful for splitting the indexed text in a list of
    tokens. The Frequency vector is also added so that the tf calculation is possible. Commas is true in case of a
    "content_id" field data containing white spaces
    """

    def __init__(self, directory: str):
        super().__init__(directory)

    @property
    def schema_type(self):
        return KEYWORD(stored=True, commas=True, vector=Frequency())

    def __str__(self):
        return "KeywordIndex"

    def __repr__(self):
        return f'KeywordIndex(directory={self.directory})'


class SearchIndex(IndexInterface):
    """
    This class implements the schema_type method: Text. By using a SimpleAnalyzer for the field, the data is kept as
    much as the original as possible
    """

    def __init__(self, directory: str):
        super().__init__(directory)

    @property
    def schema_type(self):
        return TEXT(stored=True, analyzer=SimpleAnalyzer())

    def __str__(self):
        return "SearchIndex"

    def __repr__(self):
        return f'SearchIndex(directory={self.directory})'



import setuptools
from setuptools import setup

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

with open("README.md", "r") as fh:
    long_description = fh.read()

VERSION = "0.3.1"

setup(name='clayrs',
      version=VERSION,
      license='GPL-3.0',
      author='Antonio Silletti, Elio Musacchio, Roberta Sallustio',
      install_requires=requirements,
      description='Complexly represent contents, build recommender systems, evaluate them. All in one place!',
      long_description=long_description,
      long_description_content_type="text/markdown",
      keywords=['recommender system', 'cbrs', 'evaluation', 'recsys'],
      url='https://github.com/swapUniba/ClayRS',
      include_package_data=True,
      packages=setuptools.find_packages(),
      python_requires='>=3.7',

      classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3 :: Only',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Topic :: Software Development :: Libraries',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Topic :: Software Development :: Testing :: Unit'
      ]

      )



import logging
from typing import Optional, Dict

from colorama import Fore, Back, Style


class ColoredFormatter(logging.Formatter):
    """Colored log formatter."""

    def __init__(self, *args, colors: Optional[Dict[str, str]] = None, **kwargs) -> None:
        """Initialize the formatter with specified format strings."""

        super().__init__(*args, **kwargs)

        self.colors = colors if colors else {}

    def format(self, record) -> str:
        """Format the specified record as text."""

        record.color = self.colors.get(record.levelname, '')
        record.reset = Style.RESET_ALL

        return super().format(record)


formatter = ColoredFormatter(
    "\r{color}{levelname}{reset} - {message}",
    style='{',
    colors={
        'DEBUG': Fore.CYAN,
        'INFO': Fore.RESET,
        'WARNING': Fore.YELLOW,
        'ERROR': Fore.RED,
        'CRITICAL': Fore.BLACK + Back.RED + Style.BRIGHT,
    }
)


def get_custom_logger(name: str):
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    logger = logging.getLogger(name)
    if len(logger.handlers) != 0:
        logger.handlers.clear()

    logger.addHandler(handler)
    logger.setLevel(logging.INFO)
    logger.propagate = False
    return logger



import pandas as pd
from typing import List, Any, Union, Iterator, Dict


class RepresentationContainer:
    """
    Class that stores a generic representation. This is used in the project for storing the representations and
    ids for both the field and exogenous representations of the contents. In order to store the data, the class handles
    a dataframe with 1 column ('representation') and 2 indexes ('internal_id' and 'external_id').
    The dataframe is in the following form

                                            representation
                    internal_id external_id
                    0           'test'      FieldRepresentation instance
                    1           NaN         FieldRepresentation instance
                    2           'test2'     FieldRepresentation instance


    The index 'internal_id' is used to store the id automatically assigned by the framework to the representation.
    This default id is an integer and the column will always be in the form: [0, 1, 2, 3, ...]

    The index 'external_id' is used to store the optional id that the user can assign to the representation.
    If the user didn't define an external_id for the representation it will be set to NaN and it will be
    accessible using the internal_id only.
    The 'external_id' is a string and the column will always be in the form: ['test', NaN, 'test2', 'test3', ...]

    Both 'internal_id' and 'external_id' contain unique values, meaning that there can't be duplicates in the
    columns (except for NaN)

    The column 'representation' is used to store the instances of the representations for the content,
    It can store both instances of FieldRepresentation or ExogenousRepresentation (not both at the same time obviously,
    the column has to contain FieldRepresentation or ExogenousRepresentation only)
    By using an index value either integer (referring to 'internal_id') or string (referring to 'external_id') it's
    possible to access the corresponding representation

    Args:
        external_id_list (Union[List[Union[str, None]], Union[str, None]]): list containing the user defined ids for the
            representations, the None value is used for representations the user didn't assign an id to. It's also
            possible to pass a single value instead of a list.
        representation_list (Union[List[Any], Any]): list containing the representations instances (so eiter
            FieldRepresentation or ExogenousRepresentation). It's also possible to pass a single value instead of a
            list.

    internal_id_list is not required as an argument because it will be automatically created by the class
    """

    __slots__ = ('__int_to_ext', '__ext_to_int', '__representation_container')

    def __init__(self, representation_list: Union[List[Any], Any] = None,
                 external_id_list: Union[List[Union[str, None]], Union[str, None]] = None):
        if external_id_list is None:
            external_id_list = []
        if representation_list is None:
            representation_list = []

        if not isinstance(external_id_list, list):
            external_id_list = [external_id_list]
        if not isinstance(representation_list, list):
            representation_list = [representation_list]

        if len(external_id_list) != len(representation_list):
            raise ValueError("Representation and external_id lists must have the same length")

        self.__int_to_ext = {int_id: ext_id for int_id, ext_id in enumerate(external_id_list)}
        self.__ext_to_int = {ext_id: int_id for int_id, ext_id in
                             zip(self.__int_to_ext.keys(), self.__int_to_ext.values())
                             if ext_id is not None}

        self.__representation_container = dict(zip(self.__int_to_ext.keys(), representation_list))

    def get_internal_index(self) -> List[int]:
        """
        Returns a list containing the values in the 'internal_id' index
        """
        return list(self.__int_to_ext.keys())

    def get_external_index(self) -> List[Union[str, None]]:
        """
        Returns a list containing the values in the 'external_id' index
        """
        return list(self.__int_to_ext.values())

    def get_representations(self) -> List[Any]:
        """
        Returns a list containing the values in the 'representations' column
        """
        return list(self.__representation_container.values())

    def append(self, representation: Union[List[Any], Any],
               external_id: Union[List[Union[str, None]], Union[str, None]]):
        """
        Method used to append a list of representations (or a single representation) and their list of
        external_ids (or a single external_id) to the main dataframe. In order to do so, a new dataframe is created
        for the arguments passed by the user and it will be appended to the original dataframe. The logic behind the
        creation of the dataframe is the same as the constructor, with only one difference: the internal_id index is
        generated from the original dataframe one (so that the internal_ids are consecutive).

        Args:
            external_id (Union[List[Union[str, None]], Union[str, None]]): list containing the user defined ids for the
                representations, the None value is used for representations the user didn't assign an id to. It's also
                possible to pass a single value instead of a list.
            representation (Union[List[Any], Any]): list containing the representations instances (so eiter
                FieldRepresentation or ExogenousRepresentation). It's also possible to pass a single value instead of a
                list.
        """
        if not isinstance(external_id, list):
            external_id = [external_id]
        if not isinstance(representation, list):
            representation = [representation]

        if len(representation) != len(external_id):
            raise ValueError("Representation and external_id lists must have the same length")

        if len(self.get_internal_index()) == 0:
            next_internal_id = 0
        else:
            next_internal_id = len(self.get_internal_index())

        new_int_to_ext_dict = {int_id: ext_id for int_id, ext_id in enumerate(external_id, start=next_internal_id)}

        self.__int_to_ext.update(new_int_to_ext_dict)
        self.__ext_to_int.update({ext_id: int_id for int_id, ext_id in zip(new_int_to_ext_dict.keys(),
                                                                           new_int_to_ext_dict.values())
                                  if ext_id is not None})

        self.__representation_container.update(dict(zip(new_int_to_ext_dict.keys(), representation)))

    def pop(self, id: Union[str, int]):
        """
        Remove a specific row from the dataframe identified by the external or internal id passed as an argument.
        The representation corresponding to the selected row is also returned (in case it's needed).

        Args:
            id(Union[str, int]): used to access the row to remove from the dataframe. If it is an integer, it means
                it refers to the internal_id index, if it is a string, it means that it refers to the external_id index

        Returns:
            removed_representation (Any): representation corresponding to the removed row
        """
        removed_representation = self[id]

        if isinstance(id, str):
            id_int = self.__ext_to_int[id]
            del self.__representation_container[id_int]
            del self.__int_to_ext[id_int]
            del self.__ext_to_int[id]
        else:
            id_str = self.__int_to_ext[id]
            del self.__representation_container[id]
            del self.__ext_to_int[id_str]
            del self.__int_to_ext[id]

        return removed_representation

    def __getitem__(self, item: Union[str, int]):
        """
        Access a specific representation using an index value. The index value can be either string or integer,
        if it is an integer, it means that it is referring to the 'internal_id' index, otherwise if it is a string,
        it means that it is referring to the 'external_id' index.

        Args:
            item (Union[str, int]): value used to refer to a specific representation by accessing the index columns
        """
        original_id = item
        try:
            if isinstance(item, str):
                item = self.__ext_to_int[item]

            return self.__representation_container[item]
        except KeyError:
            raise KeyError(f"Representation with id {original_id} not found!")

    def __iter__(self) -> Iterator[Dict]:
        for internal_ind, representation in self.__representation_container.items():
            yield {'internal_id': internal_ind, 'external_id': self.__int_to_ext[internal_ind],
                   'representation': representation}

    def __len__(self):
        return len(self.get_internal_index())

    def __eq__(self, other):
        return self.__representation_container == other.__representation_container

    def __str__(self):
        dataframe = pd.DataFrame({
            'internal_id': self.get_internal_index(),
            'external_id': self.get_external_index(),
            'representation': self.get_representations()
        })
        dataframe.set_index(['internal_id', 'external_id'], inplace=True)
        return str(dataframe)



from __future__ import annotations
from collections import defaultdict
from typing import List, Union, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from clayrs.content_analyzer import Content
    from clayrs.content_analyzer.field_content_production_techniques.embedding_technique.combining_technique import \
        CombiningTechnique
    from clayrs.recsys.content_based_algorithm.contents_loader import LoadedContentsDict
    from clayrs.recsys.content_based_algorithm.regressor.regressors import Regressor

from clayrs.content_analyzer.field_content_production_techniques.embedding_technique.combining_technique import \
    Centroid
from clayrs.content_analyzer.ratings_manager.ratings import Interaction
from clayrs.recsys.content_based_algorithm.exceptions import NoRatedItems, EmptyUserRatings
from clayrs.recsys.content_based_algorithm.content_based_algorithm import ContentBasedAlgorithm


class LinearPredictor(ContentBasedAlgorithm):
    """
    Class that implements recommendation through a specified linear predictor.
    It's a score prediction algorithm, so it can predict what rating a user would give to an unseen item.
    As such, it's also a ranking algorithm (it simply ranks in descending order unseen items by the predicted rating)

    Examples:

        * Interested in only a field representation, LinearRegression regressor from sklearn

        >>> from clayrs import recsys as rs
        >>> alg = rs.LinearPredictor({"Plot": 0}, rs.SkLinearRegression())

        * Interested in only a field representation, Ridge regressor from sklearn with custom parameters

        >>> alg = rs.LinearPredictor({"Plot": 0}, rs.SkRidge(alpha=0.8))

        * Interested in multiple field representations of the items, Ridge regressor from sklearn with custom
        parameters, $only_greater_eq = 2$ (Every item with rating $>= 2$ will be discarded and not considered in the
        ranking/score prediction task)

        >>> alg = rs.LinearPredictor(
        >>>                         item_field={"Plot": [0, "tfidf"],
        >>>                                     "Genre": [0, 1],
        >>>                                     "Director": "doc2vec"},
        >>>                         regressor=rs.SkRidge(alpha=0.8),
        >>>                         only_greater_eq=2)

        !!! info

            After instantiating the LinearPredictor algorithm, pass it in the initialization of
            a CBRS and the use its method to predict ratings or calculate ranking for a single user or multiple users:

            Examples:

                >>> cbrs = rs.ContentBasedRS(algorithm=alg, ...)
                >>> cbrs.fit_predict(...)
                >>> cbrs.fit_rank(...)
                >>> # ...

    Args:
        item_field: dict where the key is the name of the field
            that contains the content to use, value is the representation(s) id(s) that will be
            used for the said item. The value of a field can be a string or a list,
            use a list if you want to use multiple representations for a particular field.
        regressor: regressor that will be used. Can be one object of the `Regressor` class.
        only_greater_eq: Threshold for the ratings. Only items with rating greater or equal than the
            threshold will be considered, items with lower rating will be discarded. If None, no item will be filter out
        embedding_combiner: `CombiningTechnique` used when embeddings representation must be used but they are in a
            matrix form instead of a single vector (e.g. when WordEmbedding representations must be used you have one
            vector for each word). By default the `Centroid` of the rows of the matrix is computed
    """
    __slots__ = ('_regressor', '_labels', '_items_features', '_embedding_combiner')

    def __init__(self, item_field: dict, regressor: Regressor, only_greater_eq: float = None,
                 embedding_combiner: CombiningTechnique = Centroid()):
        super().__init__(item_field, only_greater_eq)
        self._regressor = regressor
        self._labels: Optional[list] = None
        self._items_features: Optional[list] = None
        self._embedding_combiner = embedding_combiner

    def process_rated(self, user_ratings: List[Interaction], available_loaded_items: LoadedContentsDict):
        """
        Function that extracts features from rated item and labels them.
        The extracted features will be later used to fit the classifier.

        Features and labels (in this case the rating score) will be stored in private attributes of the class.

        IF there are no rated items available locally, an exception is thrown.

        Args:
            user_ratings: List of Interaction objects for a single user
            available_loaded_items: The LoadedContents interface which contains loaded contents

        Raises:
            NoRatedItems: Exception raised when there isn't any item available locally
                rated by the user
        """
        # a list since there could be duplicate interaction (eg bootstrap partitioning)
        items_scores_dict = defaultdict(list)
        for interaction in user_ratings:
            items_scores_dict[interaction.item_id].append(interaction.score)

        items_scores_dict = dict(sorted(items_scores_dict.items()))  # sort dictionary based on key for reproducibility

        # Create list of all the available items that are useful for the user
        loaded_rated_items: List[Union[Content, None]] = available_loaded_items.get_list([item_id
                                                                                          for item_id
                                                                                          in items_scores_dict.keys()])

        # Assign label and extract features from the rated items
        labels = []
        items_features = []

        for item in loaded_rated_items:
            if item is not None:

                score_assigned = map(float, items_scores_dict[item.content_id])

                for score in score_assigned:
                    if self.threshold is None or score >= self.threshold:
                        items_features.append(self.extract_features_item(item))
                        labels.append(score)

        if len(user_ratings) == 0:
            raise EmptyUserRatings("The user selected doesn't have any ratings!")

        user_id = user_ratings[0].user_id
        if len(items_features) == 0:
            raise NoRatedItems("User {} - No rated item available locally!".format(user_id))

        self._labels = labels
        self._items_features = items_features

    def fit(self):
        """
        Fit the regressor specified in the constructor with the features and labels (rating scores)
        extracted with the process_rated() method.

        It uses private attributes to fit the classifier, so process_rated() must be called
        before this method.
        """
        # Fuse the input if there are dicts, multiple representation, etc.
        fused_features = self.fuse_representations(self._items_features, self._embedding_combiner)

        self._regressor.fit(fused_features, self._labels)

        # we delete variables used to fit since will no longer be used
        self._labels = None
        self._items_features = None

    def _common_prediction_process(self, user_ratings: List[Interaction], available_loaded_items: LoadedContentsDict,
                                   filter_list: List[str] = None):

        user_seen_items = set([interaction.item_id for interaction in user_ratings])

        # Load items to predict
        if filter_list is None:
            items_to_predict = available_loaded_items.get_list([item_id
                                                                for item_id in available_loaded_items
                                                                if item_id not in user_seen_items])
        else:
            items_to_predict = available_loaded_items.get_list(filter_list)

        # Extract features of the items to predict
        id_items_to_predict = []
        features_items_to_predict = []
        for item in items_to_predict:
            if item is not None:
                # raises AttributeError if items are not present locally
                id_items_to_predict.append(item.content_id)
                features_items_to_predict.append(self.extract_features_item(item))

        if len(id_items_to_predict) > 0:
            # Fuse the input if there are dicts, multiple representation, etc.
            fused_features_items_to_pred = self.fuse_representations(features_items_to_predict,
                                                                     self._embedding_combiner)

            score_labels = self._regressor.predict(fused_features_items_to_pred)
        else:
            score_labels = []

        return id_items_to_predict, score_labels

    def predict(self, user_ratings: List[Interaction], available_loaded_items: LoadedContentsDict,
                filter_list: List[str] = None) -> List[Interaction]:
        """
        Predicts how much a user will like unrated items.

        One can specify which items must be predicted with the filter_list parameter,
        in this case ONLY items in the filter_list will be predicted.
        One can also pass items already seen by the user with the filter_list parameter.
        Otherwise, ALL unrated items will be predicted.

        Args:
            user_ratings: List of Interaction objects for a single user
            available_loaded_items: The LoadedContents interface which contains loaded contents
            filter_list: List of the items to predict, if None all unrated items for the user will be predicted

        Returns:
            List of Interactions object where the 'score' attribute is the rating predicted by the algorithm
        """
        try:
            user_id = user_ratings[0].user_id
        except IndexError:
            raise EmptyUserRatings("The user selected doesn't have any ratings!")

        id_items_to_predict, score_labels = self._common_prediction_process(user_ratings, available_loaded_items,
                                                                            filter_list)

        # Build the output data
        pred_interaction_list = [Interaction(user_id, item_id, score)
                                 for item_id, score in zip(id_items_to_predict, score_labels)]

        return pred_interaction_list

    def rank(self, user_ratings: List[Interaction], available_loaded_items: LoadedContentsDict,
             recs_number: int = None, filter_list: List[str] = None) -> List[Interaction]:
        """
        Rank the top-n recommended items for the user. If the recs_number parameter isn't specified,
        All unrated items for the user will be ranked (or only items in the filter list, if specified).

        One can specify which items must be ranked with the `filter_list` parameter,
        in this case ONLY items in the `filter_list` parameter will be ranked.
        One can also pass items already seen by the user with the filter_list parameter.
        Otherwise, **ALL** unrated items will be ranked.

        Args:
            user_ratings: List of Interaction objects for a single user
            available_loaded_items: The LoadedContents interface which contains loaded contents
            recs_number: number of the top ranked items to return, if None all ranked items will be returned
            filter_list (list): list of the items to rank, if None all unrated items for the user will be ranked

        Returns:
            List of Interactions object in a descending order w.r.t the 'score' attribute, representing the ranking for
                a single user
        """
        try:
            user_id = user_ratings[0].user_id
        except IndexError:
            raise EmptyUserRatings("The user selected doesn't have any ratings!")

        # Predict the rating for the items and sort them in descending order
        id_items_to_predict, score_labels = self._common_prediction_process(user_ratings, available_loaded_items,
                                                                            filter_list)

        # Build the item_score dict (key is item_id, value is rank score predicted)
        # and order the keys in descending order
        item_score_dict = dict(zip(id_items_to_predict, score_labels))
        ordered_item_ids = sorted(item_score_dict, key=item_score_dict.get, reverse=True)

        # we only save the top-n items_ids corresponding to top-n recommendations
        # (if recs_number is None ordered_item_ids will contain all item_ids as the original list)
        ordered_item_ids = ordered_item_ids[:recs_number]

        # we construct the output data
        rank_interaction_list = [Interaction(user_id, item_id, item_score_dict[item_id])
                                 for item_id in ordered_item_ids]

        return rank_interaction_list

    def __str__(self):
        return "LinearPredictor"

    def __repr__(self):
        return f'LinearPredictor(item_field={self.item_field}, regressor={self._regressor}, ' \
               f'only_greater_eq={self.threshold}, embedding_combiner={self._embedding_combiner})'



from typing import List

from gensim.models.doc2vec import Doc2Vec, TaggedDocument

from clayrs.content_analyzer.embeddings.embedding_learner.embedding_learner import GensimWordEmbeddingLearner


class GensimDoc2Vec(GensimWordEmbeddingLearner):
    """
    Class that implements Doc2Vec model thanks to the Gensim library.

    If a pre-trained local Word2Vec model must be loaded, put its path in the `reference` parameter.
    Otherwise, a Word2Vec model will be trained from scratch based on the preprocessed corpus of the contents to complexly
    represent

    If you'd like to save the model once trained, set the path in the `reference` parameter and set
    `auto_save=True`. If `reference` is None, trained model won't be saved after training and will only be used to
    produce contents in the current run

    Additional parameters regarding the model itself could be passed, check [gensim documentation](https://radimrehurek.com/gensim/models/doc2vec.html)
    to see what else can be customized

    Args:
        reference: Path of the model to load/where the model trained will be saved if `auto_save=True`. If None the
            trained model won't be saved after training and will only be used to produce contents in the current run
        auto_save: If True, the model will be saved in the path specified in `reference` parameter
    """

    def __init__(self, reference: str = None, auto_save: bool = True, **kwargs):
        super().__init__(reference, auto_save, ".kv", **kwargs)

    def fit_model(self, corpus: List):
        tagged_data = [TaggedDocument(doc, [i]) for i, doc in enumerate(corpus)]
        self.model = Doc2Vec(tagged_data, **self.additional_parameters).wv

    def __str__(self):
        return "GensimDoc2Vec"

    def __repr__(self):
        return f"GensimDoc2Vec(reference={self.reference}, auto_save={self._auto_save}, " \
               f"{', '.join(f'{arg}={val}' for arg, val in self._additional_parameters.items())})"



from unittest import TestCase

import pandas as pd
import os
import lzma
import pickle

from clayrs.content_analyzer.ratings_manager.ratings import Ratings
from clayrs.recsys.graphs.nx_implementation.nx_bipartite_graphs import NXBipartiteGraph
import networkx as nx

from clayrs.recsys.graphs.graph import UserNode, ItemNode, PropertyNode

rat = pd.DataFrame.from_dict({'from_id': ["1", "1", "2", "2", "2", "3", "4", "4"],
                              'to_id': ["tt0112281", "tt0112302", "tt0112281", "tt0112346",
                                        "tt0112453", "tt0112453", "tt0112346", "tt0112453"],
                              'score': [0.8, 0.7, -0.4, 1.0, 0.4, 0.1, -0.3, 0.7]})
rat = Ratings.from_dataframe(rat)

rat_timestamp = pd.DataFrame.from_dict({'from_id': ["1", "1", "2", "2", "2", "3", "4", "4"],
                                        'to_id': ["tt0112281", "tt0112302", "tt0112281", "tt0112346",
                                                  "tt0112453", "tt0112453", "tt0112346", "tt0112453"],
                                        'score': [0.8, 0.7, -0.4, 1.0, 0.4, 0.1, -0.3, 0.7],
                                        'timestamp': ['11111', '22222', '33333', '44444',
                                                      '55555', '66666', '777777', '88888']
                                        })
rat_timestamp = Ratings.from_dataframe(rat_timestamp, timestamp_column='timestamp')


class TestNXBipartiteGraph(TestCase):

    def setUp(self) -> None:
        # graphs that will be used for testing
        self.g: NXBipartiteGraph = NXBipartiteGraph(rat)

        self.graph_custom_label: NXBipartiteGraph = NXBipartiteGraph(rat, link_label='my_label')

        self.graph_timestamp: NXBipartiteGraph = NXBipartiteGraph(rat_timestamp)

        self.empty_graph: NXBipartiteGraph = NXBipartiteGraph()

    def test_graph_creation(self):
        # test that all nodes are added to the graph
        for interaction in rat:
            self.assertTrue(UserNode(interaction.user_id) in self.g.user_nodes)
            self.assertTrue(ItemNode(interaction.item_id) in self.g.item_nodes)

            link_data = self.g.get_link_data(UserNode(interaction.user_id), ItemNode(interaction.item_id))
            self.assertIsNotNone(link_data)

            expected_data_link = {'weight': interaction.score}
            result = self.g.get_link_data(UserNode(interaction.user_id), ItemNode(interaction.item_id))

            self.assertEqual(expected_data_link, result)

    def test_graph_creation_custom_label(self):
        # test that all nodes are added to the graph and the link between them has
        # a changed label
        for interaction in rat:
            self.assertTrue(UserNode(interaction.user_id) in self.graph_custom_label.user_nodes)
            self.assertTrue(ItemNode(interaction.item_id) in self.graph_custom_label.item_nodes)

            link_data = self.graph_custom_label.get_link_data(UserNode(interaction.user_id),
                                                              ItemNode(interaction.item_id))
            self.assertIsNotNone(link_data)

            expected_data_link = {'label': 'my_label', 'weight': interaction.score}
            result = self.graph_custom_label.get_link_data(UserNode(interaction.user_id), ItemNode(interaction.item_id))

            self.assertEqual(expected_data_link, result)

    def test_graph_creation_w_timestamp(self):
        for interaction in rat_timestamp:
            expected_data_link = {'weight': interaction.score, 'timestamp': interaction.timestamp}
            result = self.graph_timestamp.get_link_data(UserNode(interaction.user_id), ItemNode(interaction.item_id))

            self.assertEqual(expected_data_link, result)

    def test_graph_creation_empty(self):
        self.assertTrue(len(self.empty_graph.user_nodes) == 0)
        self.assertTrue(len(self.empty_graph.item_nodes) == 0)

    def test_add_user(self):
        # Add 'user' node
        self.assertFalse(UserNode('u0') in self.g.user_nodes)
        self.assertFalse(self.g.node_exists(UserNode('u0')))
        self.g.add_node(UserNode('u0'))
        self.assertTrue(self.g.node_exists(UserNode('u0')))
        self.assertTrue(UserNode('u0') in self.g.user_nodes)

        # Add a list of 'user' nodes
        list_nodes = [UserNode('u1_list'), UserNode('u2_list'), UserNode('u3_list')]
        self.g.add_node(list_nodes)
        for n in list_nodes:
            self.assertTrue(self.g.node_exists(n))
            self.assertTrue(n in self.g.user_nodes)

        # Add 'user' node with same id of an already existent
        # 'item' node, verify that both exists as different entities
        self.g.add_node(ItemNode('i0'))
        self.g.node_exists(ItemNode('i0'))
        self.assertTrue(ItemNode('i0') in self.g.item_nodes)
        self.g.add_node(UserNode('i0'))
        self.assertTrue(self.g.node_exists(UserNode('i0')))
        self.assertTrue(self.g.node_exists(ItemNode('i0')))
        self.assertTrue(UserNode('i0') in self.g.user_nodes)
        self.assertTrue(ItemNode('i0') in self.g.item_nodes)

    def test_add_item(self):
        # Add 'item' node
        self.assertFalse(ItemNode('i0') in self.g.item_nodes)
        self.g.add_node(ItemNode('i0'))
        self.assertTrue(self.g.node_exists(ItemNode('i0')))
        self.assertTrue(ItemNode('i0') in self.g.item_nodes)

        # Add a list of 'item' nodes
        list_nodes = [ItemNode('i1_list'), ItemNode('i2_list'), ItemNode('i3_list')]
        self.g.add_node(list_nodes)
        for n in list_nodes:
            self.assertTrue(self.g.node_exists(n))
            self.assertTrue(n in self.g.item_nodes)

        # Add 'item' node with same id of an already existent
        # 'user' node, verify that both exists as different entities
        self.g.add_node(UserNode('u0'))
        self.assertTrue(self.g.node_exists(UserNode('u0')))
        self.assertTrue(UserNode('u0') in self.g.user_nodes)
        self.g.add_node(ItemNode('u0'))
        self.assertTrue(self.g.node_exists(UserNode('u0')))
        self.assertTrue(self.g.node_exists(ItemNode('u0')))
        self.assertTrue(UserNode('u0') in self.g.user_nodes)
        self.assertTrue(ItemNode('u0') in self.g.item_nodes)

    def test_add_raise_error(self):

        # test add a node which is not item node nor user node
        with self.assertRaises(ValueError):
            self.g.add_node(PropertyNode('prop'))

        # test add a list of node in which a node is not a user or item node
        with self.assertRaises(ValueError):
            self.g.add_node([UserNode('u'), ItemNode('i'), PropertyNode('p')])

    def test_add_link_user_item_existent(self):
        # Link existent 'user' node to an existent 'item' node
        user_node = UserNode('u0')
        item_node = ItemNode('Tenet')
        self.g.add_node(user_node)
        self.g.add_node(item_node)
        self.assertIsNone(self.g.get_link_data(user_node, item_node))
        self.g.add_link(user_node, item_node, 0.5)
        expected = {'weight': 0.5}
        result = self.g.get_link_data(user_node, item_node)
        self.assertEqual(expected, result)

        # Link list of existent 'user' nodes to a list of existent 'item' nodes
        user_list = [UserNode('u1_list'), UserNode('u1_list'), UserNode('u3_list')]
        self.g.add_node(user_list)
        items_list = [ItemNode('i1_list'), ItemNode('i2_list'), ItemNode('i3_list')]
        self.g.add_node(items_list)
        self.g.add_link(user_list, items_list, 0.5)
        for user, item in zip(user_list, items_list):
            result = self.g.get_link_data(user, item)
            expected = {'weight': 0.5}

            self.assertEqual(expected, result)

        # Link existent 'item' node to an existent 'user' node
        user_node = UserNode('1')
        item_node = ItemNode('Tenet')
        self.g.add_node(item_node)
        self.g.add_node(user_node)
        self.assertIsNone(self.g.get_link_data(item_node, user_node))
        self.g.add_link(item_node, user_node, 0.5)
        self.assertFalse(self.g.node_exists(UserNode('Tenet')))
        self.assertFalse(self.g.node_exists(ItemNode('u1')))
        self.assertIsNotNone(self.g.get_link_data(item_node, user_node))

        # Link list of existent 'item' nodes to a list of existent 'user' nodes
        items_list = [ItemNode('i1_list'), ItemNode('i2_list'), ItemNode('i3_list')]
        self.g.add_node(items_list)
        users_list = [UserNode('u1_list'), UserNode('u2_list'), UserNode('u3_list')]
        self.g.add_node(users_list)
        self.g.add_link(items_list, users_list, 0.5)
        for item, user in zip(items_list, users_list):
            result = self.g.get_link_data(item, user)
            expected = {'weight': 0.5}

            self.assertEqual(expected, result)

    def test_add_link_user_item_non_existent(self):

        # Link non-existent 'user' node and non-existent 'item' node,
        # so both nodes are created and then linked
        user_new = UserNode('u_new')
        item_new = ItemNode('i_new')
        self.assertFalse(self.g.node_exists(user_new))
        self.assertFalse(self.g.node_exists(item_new))
        self.g.add_link(user_new, item_new, 0.5)
        self.assertTrue(self.g.node_exists(user_new))
        self.assertTrue(self.g.node_exists(item_new))
        self.assertIsNotNone(self.g.get_link_data(user_new, item_new))

        # Link non-existent 'user' node list and non-existent 'item' node list,
        # so all nodes of the two lists are created and then linked
        user_new_list = [UserNode('u_new_new1'), UserNode('u_new_new2')]
        item_new_list = [ItemNode('i_new_new1'), ItemNode('i_new_new2')]
        for user in user_new_list:
            self.assertFalse(self.g.node_exists(user))
        for item in item_new_list:
            self.assertFalse(self.g.node_exists(item))

        self.g.add_link(user_new_list, item_new_list, 0.5)

        for user, item in zip(user_new_list, item_new_list):
            result = self.g.get_link_data(user, item)
            expected = {'weight': 0.5}

            self.assertEqual(expected, result)

    def test_add_link_raise_error(self):
        # test link property node in a bipartite graph
        with self.assertRaises(ValueError):
            self.g.add_link(PropertyNode('prop'), ItemNode('item'))

    def test_add_link_custom_parameters(self):
        # Link non-existent 'user' node and non-existent 'item' node
        # with no parameters
        user_new = UserNode('u_new')
        item_new = ItemNode('i_new')
        self.assertFalse(self.g.node_exists(user_new))
        self.assertFalse(self.g.node_exists(item_new))
        self.g.add_link(user_new, item_new, label=None)
        self.assertTrue(self.g.node_exists(user_new))
        self.assertTrue(self.g.node_exists(item_new))
        expected = {}
        result = self.g.get_link_data(user_new, item_new)

        self.assertEqual(expected, result)

        # Link non-existent 'user' node and non-existent 'item' node
        # with all custom parameters
        user_new = UserNode('u_new1')
        item_new = ItemNode('i_new1')
        self.assertFalse(self.g.node_exists(user_new))
        self.assertFalse(self.g.node_exists(item_new))
        self.g.add_link(user_new, item_new, weight=0.5, label='my_label', timestamp='001122')
        self.assertTrue(self.g.node_exists(user_new))
        self.assertTrue(self.g.node_exists(item_new))
        expected = {'label': 'my_label', 'weight': 0.5, 'timestamp': '001122'}
        result = self.g.get_link_data(user_new, item_new)

        self.assertEqual(expected, result)

    def test_remove_link(self):
        # test remove link between existent nodes
        user_new = UserNode('u_new')
        item_new = ItemNode('i_new')
        self.g.add_link(user_new, item_new, label='test_link')

        # test that the link is present
        expected = {'label': 'test_link'}
        result = self.g.get_link_data(user_new, item_new)
        self.assertEqual(expected, result)

        # test that the link is removed
        self.g.remove_link(user_new, item_new)
        self.assertIsNone(self.g.get_link_data(user_new, item_new))

        # test remove link between non existent nodes
        result = self.g.remove_link(UserNode('non_existent'), ItemNode('non_existent1'))
        self.assertIsNone(result)

    def test_pred_succ(self):
        # Get predecessors of a node
        self.g.add_link(UserNode('u0'), ItemNode('i0'), 0.5)
        self.g.add_link(UserNode('u1'), ItemNode('i0'), 0.5)
        result = self.g.get_predecessors(ItemNode('i0'))
        expected = [UserNode('u0'), UserNode('u1')]
        self.assertEqual(expected, result)

        # Get predecessors of a non-existent node
        with self.assertRaises(TypeError):
            self.g.get_predecessors(ItemNode('non_existent'))

        # Get successors of a node
        self.g.add_link(UserNode('u0'), ItemNode('i0'), 0.5)
        self.g.add_link(UserNode('u0'), ItemNode('i1'), 0.5)
        result = self.g.get_successors(UserNode('u0'))
        expected = [ItemNode('i0'), ItemNode('i1')]
        self.assertEqual(expected, result)

        # Get successors of a non-existent node
        with self.assertRaises(TypeError):
            self.g.get_successors(UserNode('non_existent'))

    def test_metrics(self):
        # We calculate some metrics, simple assert to make sure they are
        # calculated
        self.assertGreater(len(self.g.degree_centrality()), 0)
        self.assertGreater(len(self.g.closeness_centrality()), 0)
        self.assertGreater(len(self.g.dispersion()), 0)

    def test_to_networkx(self):
        # Simple assert just to test the _graph method
        self.assertIsInstance(self.g.to_networkx(), nx.DiGraph)

    def test_remove_node(self):
        # remove a single node
        node = UserNode("to_remove")
        self.g.add_node(node)
        self.assertTrue(node in self.g.user_nodes)
        self.g.remove_node(node)
        self.assertFalse(node in self.g.user_nodes)

        # remove multiple nodes
        nodes = [UserNode("to_remove"), ItemNode("to_remove")]
        self.g.add_node(nodes)
        self.assertTrue(nodes[0] in self.g.user_nodes)
        self.assertTrue(nodes[1] in self.g.item_nodes)
        self.g.remove_node(nodes)
        self.assertFalse(nodes[0] in self.g.user_nodes)
        self.assertFalse(nodes[1] in self.g.item_nodes)

        # remove not existent node
        user_nodes_before_removal = self.g.user_nodes
        self.g.remove_node(UserNode("not existent"))
        user_nodes_after_removal = self.g.user_nodes
        self.assertEqual(user_nodes_before_removal, user_nodes_after_removal)

    #
    # def test_convert_to_dataframe(self):
    #     converted_df = self.g.convert_to_dataframe()
    #     self.assertNotIn('label', converted_df.columns)
    #     for user, item in zip(converted_df['from_id'], converted_df['to_id']):
    #         self.assertIsInstance(user, Node)
    #         self.assertIsInstance(item, Node)
    #
    #     result = np.sort(converted_df, axis=0)
    #     expected = np.sort(self.df, axis=0)
    #     self.assertTrue(np.array_equal(expected, result))
    #
    #     converted_df = self.g.convert_to_dataframe(only_values=True, with_label=True)
    #     self.assertIn('label', converted_df.columns)
    #     for user, item in zip(converted_df['from_id'], converted_df['to_id']):
    #         self.assertNotIsInstance(user, Node)
    #         self.assertNotIsInstance(item, Node)
    #
    #     converted_df = converted_df[['from_id', 'to_id', 'score']]
    #     result = np.sort(converted_df, axis=0)
    #     expected = np.sort(self.df, axis=0)
    #     self.assertTrue(np.array_equal(expected, result))
    #
    def test_copy(self):
        copy = self.g.copy()
        self.assertTrue(copy == self.g)

        copy.add_node(UserNode('prova'))

        self.assertFalse(copy == self.g)

    def test_serialize(self):

        self.g.serialize('./test_serialize', 'test_graph')

        with lzma.open(os.path.join('./test_serialize/test_graph.xz'), 'rb') as graph_file:
            graph = pickle.load(graph_file)

        self.assertEqual(self.g, graph)



from unittest import TestCase

import numpy as np

from clayrs.content_analyzer.content_representation.content import Content, PropertiesDict, FeaturesBagField
from clayrs.content_analyzer.content_representation.representation_container import RepresentationContainer


class TestContent(TestCase):
    def test_append_remove_field(self):
        """
        Tests for append, remove and get methods of the content's field instances
        """
        features_bag_vector = np.array([1.0, 0])
        features_bag_pos_features = [(0, 'word_present'), (1, 'word_not_present')]

        content_field_repr = FeaturesBagField(features_bag_vector, features_bag_pos_features)
        content_field = RepresentationContainer()
        content_field.append(content_field_repr, "test_1")
        content1 = Content("001")
        content1.append_field("test_field", content_field)

        content2 = Content("002")
        content2.append_field("test_field", content_field)
        content_field_repr = FeaturesBagField(features_bag_vector, features_bag_pos_features)
        content_field2 = RepresentationContainer()
        content_field2.append(content_field_repr, "test_1")
        content2.append_field("test_field2", content_field2)
        content2.remove_field("test_field2")
        self.assertEqual(content1.field_dict, content2.field_dict)
        self.assertEqual(content1.get_field("test_field"), content2.get_field("test_field"))

    def test_append_remove_field_repr(self):
        """
        Tests for append, remove and get methods of the content's fields' representation instances
        """
        features_bag_vector = np.array([1.0, 0])
        features_bag_pos_features = [(0, 'word_present'), (1, 'word_not_present')]

        # test append_field_repr when field already existent
        content_field_repr_1 = FeaturesBagField(features_bag_vector, features_bag_pos_features)
        content_field = RepresentationContainer(content_field_repr_1, "test_1")
        content = Content("001")
        content.append_field("test_field", content_field)

        # test append_field_repr when field not existent
        content_field_repr_2 = FeaturesBagField(features_bag_vector, features_bag_pos_features)
        content.append_field_representation("test_field_2", content_field_repr_2, "test_2")
        self.assertEqual(content.get_field_representation("test_field_2", "test_2"), content_field_repr_2)
        self.assertEqual(len(content.get_field("test_field_2")), 1)

        # test append_field_repr with list of repr
        content_field_repr_3_first = FeaturesBagField(features_bag_vector, features_bag_pos_features)
        content_field_repr_3_second = FeaturesBagField(features_bag_vector, features_bag_pos_features)
        content.append_field_representation("test_field_3", [content_field_repr_3_first, content_field_repr_3_second],
                                            ["test_3_first", "test_3_second"])
        self.assertEqual(content.get_field_representation("test_field_3", "test_3_first"), content_field_repr_3_first)
        self.assertEqual(content.get_field_representation("test_field_3", "test_3_second"), content_field_repr_3_second)
        self.assertEqual(len(content.get_field("test_field_3")), 2)

        # test append_field_repr with list of repr and no id specified
        content_field_repr_4_first = FeaturesBagField(features_bag_vector, features_bag_pos_features)
        content_field_repr_4_second = FeaturesBagField(features_bag_vector, features_bag_pos_features)
        content.append_field_representation("test_field_4", [content_field_repr_4_first, content_field_repr_4_second])
        self.assertEqual(content.get_field_representation("test_field_4", 0), content_field_repr_4_first)
        self.assertEqual(content.get_field_representation("test_field_4", 1), content_field_repr_4_second)
        self.assertEqual(len(content.get_field("test_field_4")), 2)

        # test remove
        content.remove_field_representation("test_field_2", "test_2")
        self.assertEqual(len(content.get_field("test_field_2")), 0)

    def test_append_remove_exo(self):
        """
        Tests for append, remove and get methods of the content's exogenous instances
        """
        exo_features = dict()
        exo_features["test_key"] = "test_value"

        content_exo_repr = PropertiesDict(exo_features)
        content_exo_repr2 = PropertiesDict({"test_key2": 'test_value2'})

        content1 = Content("001")
        content1.append_exogenous_representation(content_exo_repr, "test_exo")

        content2 = Content("002")
        content2.append_exogenous_representation(content_exo_repr, "test_exo")
        content_exo_repr = PropertiesDict(exo_features)
        content2.append_exogenous_representation(content_exo_repr, "test_exo2")
        content2.remove_exogenous_representation("test_exo2")
        self.assertEqual(content1.exogenous_rep_container, content2.exogenous_rep_container)
        self.assertEqual(content1.get_exogenous_representation("test_exo"),
                         content2.get_exogenous_representation("test_exo"))

        # test append list of representations
        content3 = Content("003")

        content3.append_exogenous_representation([content_exo_repr, content_exo_repr2], ["id1", "id2"])
        self.assertEqual(len(content3.exogenous_rep_container), 2)
        self.assertEqual(content3.get_exogenous_representation("id1").value, content_exo_repr.value)
        self.assertEqual(content3.get_exogenous_representation("id2").value, content_exo_repr2.value)

        # test append list of representations without id
        content4 = Content("004")
        content4.append_exogenous_representation([content_exo_repr, content_exo_repr2])
        self.assertEqual(len(content3.exogenous_rep_container), 2)
        self.assertEqual(content3.get_exogenous_representation(0).value, content_exo_repr.value)
        self.assertEqual(content3.get_exogenous_representation(1).value, content_exo_repr2.value)



import unittest
from collections import Counter
from unittest import TestCase
import pandas as pd

from clayrs.content_analyzer import Ratings
from clayrs.evaluation.utils import get_most_popular_items, pop_ratio_by_user, get_avg_pop, get_item_popularity


class TestUtils(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        df = pd.DataFrame({
            'from_id': ['u1', 'u1', 'u2', 'u2', 'u3', 'u3', 'u4', 'u4', 'u5', 'u5'],
            'to_id': ['i1', 'i2', 'i1', 'i50', 'i1', 'i2', 'i1', 'i50', 'i2', 'i70'],
            'score': [5, 2.5, 3, 4, 5, 1.2, 5, 4, 4.5, 3]
        })
        cls.custom_rat = Ratings.from_dataframe(df)

    def test_get_popularity(self):

        # n_frequency / n_users
        expected_pop_by_item = {'i1': 4 / 5, 'i2': 3 / 5, 'i50': 2 / 5, 'i70': 1 / 5}
        result_pop_by_item = get_item_popularity(self.custom_rat)

        self.assertCountEqual(expected_pop_by_item, result_pop_by_item)

    def test_popular_items(self):
        # there are 4 unique items, default percentage = 0.2, will return 1 most popular item
        pop_by_item = get_item_popularity(self.custom_rat)
        result = get_most_popular_items(pop_by_item)
        expected = {'i1'}

        self.assertEqual(expected, result)

        # there are 4 unique items, percentage = 0.5, will return 2 most popular item
        result = get_most_popular_items(pop_by_item, 0.5)
        expected = {'i1', 'i2'}

        self.assertEqual(expected, result)

    def test_pop_ratio_by_user(self):
        most_popular_items = {'i1'}
        result = pop_ratio_by_user(self.custom_rat, most_popular_items)

        # Expected popularity ratio is:
        # the number of item rated by the user that are in most_popular_items / n_rated items by the user

        expected = {'u1': 1/2, 'u2': 1/2, 'u3': 1/2, 'u4': 1/2, 'u5': 0}

        self.assertTrue(expected, result)

    def test_get_avg_pop(self):

        counter_popularity = Counter(self.custom_rat.item_id_column)

        # Expected result are item_popularity rated by user / n_item rated by user
        # item_popularity is the number of occurrences of the item in the 'to_id' column

        u1_items = [interaction.item_id for interaction in self.custom_rat.get_user_interactions('u1')]
        result_u1 = get_avg_pop(u1_items, counter_popularity)
        expected_u1 = (counter_popularity['i1'] + counter_popularity['i2']) / 2
        self.assertAlmostEqual(expected_u1, result_u1)

        u2_items = [interaction.item_id for interaction in self.custom_rat.get_user_interactions('u2')]
        result_u2 = get_avg_pop(u2_items, counter_popularity)
        expected_u2 = (counter_popularity['i1'] + counter_popularity['i50']) / 2
        self.assertAlmostEqual(expected_u2, result_u2)

        u3_items = [interaction.item_id for interaction in self.custom_rat.get_user_interactions('u3')]
        result_u3 = get_avg_pop(u3_items, counter_popularity)
        expected_u3 = (counter_popularity['i1'] + counter_popularity['i2']) / 2
        self.assertAlmostEqual(expected_u3, result_u3)

        u4_items = [interaction.item_id for interaction in self.custom_rat.get_user_interactions('u4')]
        result_u4 = get_avg_pop(u4_items, counter_popularity)
        expected_u4 = (counter_popularity['i1'] + counter_popularity['i50']) / 2
        self.assertAlmostEqual(expected_u4, result_u4)

        u5_items = [interaction.item_id for interaction in self.custom_rat.get_user_interactions('u5')]
        result_u5 = get_avg_pop(u5_items, counter_popularity)
        expected_u5 = (counter_popularity['i2'] + counter_popularity['i70']) / 2
        self.assertAlmostEqual(expected_u5, result_u5)


if __name__ == '__main__':
    unittest.main()



from functools import wraps


def Handler_ScoreNotFloat(func):
    """
    Handler that catches the above exception.

    Tries to run the functions normally, if one of the above exceptions is caught then it must return
    an empty frame for the user since predictions can't be calculated for it.
    """
    @wraps(func)
    def Inner_Function(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ValueError:
            raise ValueError("The 'score' column must contains numbers!\n"
                             "Try the same column with a score processor or change column!")

    return Inner_Function



from __future__ import annotations
from typing import List, Set, TYPE_CHECKING

if TYPE_CHECKING:
    from clayrs.content_analyzer.ratings_manager.ratings import Interaction, Ratings
    from clayrs.recsys.graphs import NXBipartiteGraph
    from clayrs.recsys.methodology import Methodology

from clayrs.recsys.content_based_algorithm.exceptions import NotPredictionAlg
from clayrs.recsys.graph_based_algorithm.graph_based_algorithm import GraphBasedAlgorithm
from clayrs.recsys.methodology import TestRatingsMethodology


class PageRank(GraphBasedAlgorithm):
    """
    Abstract class that contains the main methods and attributes for any PageRank algorithm.

    Every PageRank algorithm can be `personalized`, in this case the PageRank will be calculated with a personalization
    vector made by items in the user profile weighted by the score given to them.

    Args:
        personalized (bool): boolean value that specifies if the page rank must be calculated considering the user
            profile as personalization vector. Default is False
    """

    def __init__(self, personalized: bool = False):

        self._personalized = personalized
        super().__init__()

    def predict(self, all_users: Set[str], graph: NXBipartiteGraph, test_set: Ratings,
                methodology: Methodology = TestRatingsMethodology(),
                num_cpus: int = 1) -> List[Interaction]:
        """
        PageRank is not a Prediction Score Algorithm, so if this method is called,
        a NotPredictionAlg exception is raised

        Raises:
            NotPredictionAlg: exception raised since the PageRank algorithm is not a score prediction algorithm
        """
        raise NotPredictionAlg("PageRank is not a Score Prediction Algorithm!")



class UserSkipAlgFit(Exception):
    """
    Super class for exception related to the fit of a single user. If one of the exception happens, the algorithm
    can't be fitted for the user, therefore will be skipped
    """
    pass


class OnlyPositiveItems(UserSkipAlgFit):
    """
    Exception to raise when there's only positive items available locally for the user
    """
    pass


class OnlyNegativeItems(UserSkipAlgFit):
    """
    Exception to raise when there's only negative items available locally for the user
    """
    pass


class NoRatedItems(UserSkipAlgFit):
    """
    Exception to raise when there's no item available locally for the user
    """
    pass


class EmptyUserRatings(UserSkipAlgFit):
    """
    Exception to raise when the user ratings is empty
    """
    pass


class NotRankingAlg(Exception):
    """
    Exception to raise when the algorithm is not a ranking algorithm, but it is asked to rank
    """
    pass


class NotPredictionAlg(Exception):
    """
    Exception to raise when the algorithm is not a prediction algorithm, but it is asked to predict
    """
    pass


class NotFittedAlg(Exception):
    """
    Exception to raise when the algorithm has not been fitted
    """
    pass



from typing import List


def id_values_merger(id_values: List[str]):
    """
    This function is used to compact a list of ids into a unique string. This can be useful when
    there is content whose id is composed by values coming from more than one field.

    Args:
        id_values (List<str>): List containing one or more ids

    Returns:
        id_merged (str): String in which the values contained in the list given in input are
        merged
    """
    if type(id_values) == str or type(id_values) == int:
        return str(id_values)
    elif type(id_values) == list:
        id_merged = ""
        for i in range(len(id_values)):
            id_merged += str(id_values[i])
            if i != len(id_values) - 1:
                id_merged += "_"
        return id_merged
    else:
        raise TypeError("id must be an integer, a string or a list of strings and/or integer")


def id_merger(raw_content: dict, field_list: List[str]) -> str:
    """
    Function that creates the list of ids and then calls id_values_merger to create a unique id
    """
    id_values = []
    for field_name in field_list:
        id_values.append(raw_content[field_name])

    return id_values_merger(id_values)



from __future__ import annotations
from typing import List, Union, Dict, TYPE_CHECKING

from clayrs.utils.const import logger

if TYPE_CHECKING:
    from clayrs.content_analyzer import Ratings, Content
    from clayrs.recsys.graphs.graph import Node

from clayrs.recsys.content_based_algorithm.contents_loader import LoadedContentsDict
from clayrs.recsys.graphs.nx_implementation import NXTripartiteGraph
from clayrs.recsys.graphs.graph import FullDiGraph, UserNode, PropertyNode
from clayrs.utils.context_managers import get_progbar


# Multiple Inheritance so that we will use NXTripartite as an interface (we only use its methods)
# and FullGraph as its proper father class (we'll call its __init__)
class NXFullGraph(NXTripartiteGraph, FullDiGraph):
    """
    Class that implements a Full graph through networkx library.

    !!! info

        A *Full Graph* is a graph which doesn't impose any particular restriction

    It creates a graph from an initial Rating object.

    Consider the following matrix representation of the Rating object
    ```
        +------+-----------+-------+
        | User |   Item    | Score |
        +------+-----------+-------+
        | u1   | Tenet     |     4 |
        | u2   | Inception |     5 |
        | ...  | ...       |   ... |
        +------+-----------+-------+
    ```

    The graph will be created with the following interactions:

    ```
                 4
            u1 -----> Tenet
                 5
            u2 -----> Inception
    ```

    where `u1` and `u2` become *User nodes* and `Tenet` and `Inception` become *Item nodes*,
    with the edge weighted depending on the score given

    If the `link_label` parameter is specified, then each link between users and items will be labeled with the label
    specified (e.g. `link_label='score'`):

    ```
            (4, 'score')
        u1 -------------> Tenet
            (5, 'score')
        u2 -------------> Inception
    ```

    Then the framework tries to load 'Tenet' and 'Inception' from the `item_contents_dir` and 'u1' and 'u2' from
    `user_contents_dir` if they are specified and if it succeeds, adds in the graph their loaded properties as
    specified in the `item_exo_properties` parameter and `user_exo_properties`.

    !!! info "Load exogenous properties"

        In order to load properties in the graph, we must specify where users (and/or) items are serialized and ***which
        properties to add*** (the following is the same for *item_exo_properties*):

        *   If *user_exo_properties* is specified as a **set**, then the graph will try to load **all properties**
        from **said exogenous representation**

        ```python
        {'my_exo_id'}
        ```

        *   If *user_exo_properties* is specified as a **dict**, then the graph will try to load **said properties**
        from **said exogenous representation**

        ```python
        {'my_exo_id': ['my_prop1', 'my_prop2']]}
        ```

    Args:
        source_frame: The initial Ratings object needed to create the graph
        item_exo_properties: Set or Dict which contains representations to load from items. Use a `Set` if you want
            to load all properties from specific representations, or use a `Dict` if you want to choose which properties
            to load from specific representations
        item_contents_dir: The path containing items serialized with the Content Analyzer
        user_exo_properties: Set or Dict which contains representations to load from items. Use a `Set` if you want
            to load all properties from specific representations, or use a `Dict` if you want to choose which properties
            to load from specific representations
        user_contents_dir: The path containing users serialized with the Content Analyzer
        link_label: If specified, each link will be labeled with the given label. Default is None

    """

    def __init__(self, source_frame: Ratings = None,
                 item_exo_properties: Union[Dict, set] = None,
                 item_contents_dir: str = None,
                 user_exo_properties: Union[Dict, set] = None,
                 user_contents_dir: str = None,
                 link_label: str = None):

        NXTripartiteGraph.__init__(self, source_frame, item_exo_properties, item_contents_dir, link_label)

        if user_exo_properties and not user_contents_dir:
            logger.warning("`user_exo_properties` parameter set but `user_contents_dir` is None! "
                           "No property will be loaded")
        elif not user_exo_properties and user_contents_dir:
            logger.warning("`user_contents_dir` parameter set but `user_exo_properties` is None! "
                           "No property will be loaded")

        if source_frame is not None and user_contents_dir is not None and user_exo_properties is not None:
            self.add_node_with_prop([UserNode(user_id) for user_id in set(source_frame.user_id_column)],
                                    user_exo_properties,
                                    user_contents_dir)

    def add_node(self, node: Union[Node, List[Node]]):
        """
        Adds one or multiple Node objects to the graph.
        Since this is a Full Graph, any category of node is allowed

        No duplicates are allowed, but different category nodes with same id are (e.g. `ItemNode('1')` and
        `UserNode('1')`)

        Args:
            node: Node(s) object(s) that needs to be added to the graph
        """
        if not isinstance(node, list):
            node = [node]

        self._graph.add_nodes_from(node)

    def add_link(self, start_node: Union[Node, List[Node]], final_node: Union[Node, List[Node]],
                 weight: float = None, label: str = None, timestamp: str = None):
        """
        Creates a weighted link connecting the 'start_node' to the 'final_node'
        Both nodes must be present in the graph before calling this method

        'weight' and 'label' are optional parameters, if not specified default values
        will be used.

        Args:
            start_node (object): starting node of the link
            final_node (object): ending node of the link
            weight (float): weight of the link, default is 0.5
            label (str): label of the link, default is 'score_label'
        """
        if not isinstance(start_node, list):
            start_node = [start_node]

        if not isinstance(final_node, list):
            final_node = [final_node]

        self.add_node(start_node)
        self.add_node(final_node)

        not_none_dict = {}
        if label is not None:
            not_none_dict['label'] = label
        if weight is not None:
            not_none_dict['weight'] = weight
        if timestamp is not None:
            not_none_dict['timestamp'] = timestamp

        self._graph.add_edges_from(zip(start_node, final_node),
                                   **not_none_dict)

    def add_node_with_prop(self, node: Union[Node, List[Node]], exo_properties: Union[Dict, set],
                           contents_dir: str,
                           content_filename: Union[str, List[str]] = None):
        """
        Adds one or multiple Node objects and its/their properties to the graph
        Since this is a Full Graph, no restriction are imposed and you can add any category of node together with its
        properties.

        In order to load properties in the graph, we must specify where contents are serialized and ***which
        properties to add*** (the following is the same for *item_exo_properties*):

        *   If *exo_properties* is specified as a **set**, then the graph will try to load **all properties**
        from **said exogenous representation**

        ```python
        {'my_exo_id'}
        ```

        *   If *exo_properties* is specified as a **dict**, then the graph will try to load **said properties**
        from **said exogenous representation**

        ```python
        {'my_exo_id': ['my_prop1', 'my_prop2']]}
        ```

        In case you want your node to have a different id from serialized contents, via the `content_filename` parameter
        you can specify what is the filename of the node that you are adding, e.g.

        ```
        item_to_add = ItemNode('different_id')

        # content_filename is 'item_serialized_1.xz'

        graph.add_node_with_prop(item_to_add, ..., content_filename='item_serialized_1')
        ```

        In case you are adding a list of nodes, you can specify the filename for each node in the list.

        Args:
            node: Node(s) object(s) that needs to be added to the graph along with their properties
            exo_properties: Set or Dict which contains representations to load from items. Use a `Set` if you want
                to load all properties from specific representations, or use a `Dict` if you want to choose which
                properties to load from specific representations
            contents_dir: The path containing items serialized with the Content Analyzer
            content_filename: Filename(s) of the node(s) to add

        Raises:
            ValueError: Exception raised when one of the node to add to the graph with their properties is not
                an ItemNode
        """
        def node_prop_link_generator():
            for n, id in zip(progbar, content_filename):
                item: Content = loaded_items.get(id)

                if item is not None:
                    exo_props = self._get_exo_props(exo_properties, item)

                    single_item_prop_edges = [(n,
                                               PropertyNode(prop_dict[prop]),
                                               {'label': prop})
                                              for prop_dict in exo_props for prop in prop_dict]
                else:
                    single_item_prop_edges = []

                yield from single_item_prop_edges

        if not isinstance(node, list):
            node = [node]

        if isinstance(exo_properties, set):
            exo_properties = dict.fromkeys(exo_properties, None)

        if content_filename is None:
            content_filename = [n.value for n in node]

        if not isinstance(content_filename, list):
            content_filename = [content_filename]

        loaded_items = LoadedContentsDict(contents_dir, contents_to_load=set(content_filename))
        with get_progbar(node) as progbar:
            progbar.set_description("Creating Node->Properties links")

            self._graph.add_edges_from((tuple_to_add for tuple_to_add in node_prop_link_generator()))

    def __str__(self):
        return "NXFullGraph"

    def __repr__(self):
        return str(self)



import inspect
from abc import ABC
from typing import Union, Any

import numpy as np
from scipy import sparse

from sklearn.linear_model import LinearRegression, BayesianRidge, Ridge, SGDRegressor, ARDRegression, \
    HuberRegressor, PassiveAggressiveRegressor
from sklearn.linear_model._stochastic_gradient import DEFAULT_EPSILON

from clayrs.utils.automatic_methods import autorepr


class Regressor(ABC):
    """
    Abstract class for Regressors
    """

    def __init__(self, model, currentframe):
        self.__model = model

        self._repr_string = autorepr(self, currentframe)

    @property
    def model(self):
        return self.__model

    def fit(self, X: list, Y: list = None):
        """
        Fit the instantiated regressor.

        Training data (X) is in the form:

            `X = [ (merged) representation for item 1, (merged) representation for item 2, ...]`

        where every item is a representation for an item (can be a merged one in case multiple representations
        are chosen)

        Target data (Y) is in the form:
            `Y = [1.2, 4.0, ... ]`

        where all labels are the numeric score given by the user

        Args:
            X: list containing Training data.
            Y: list containing Training targets.
        """
        self.__model = self.__model.fit(X, Y)

    def predict(self, X_pred: list):
        """
        Predicts the score for every item in X_pred.
        It uses the method predict() from sklearn of the instantiated regressor

        It's in the form:
            `X_pred = [ (merged) representation for item 1, (merged) representation for item 2, ...]`

        where every item is a representation for an item (can be a merged one in case multiple representations
        are chosen)

        Args:
            X_pred: list containing data to predict.
        """
        return self.__model.predict(X_pred)

    def __repr__(self):
        return self._repr_string


class SkLinearRegression(Regressor):
    """
    Class that implements the LinearRegression regressor from sklearn.
    The parameters one could pass are the same ones you would pass instantiating
    the regressor LinearRegression directly from sklearn.

    Sklearn documentation: [here](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LinearRegression.html)
    """

    def __init__(self, *, fit_intercept: Any = True,
                 normalize: Any = "deprecated",
                 copy_X: Any = True,
                 n_jobs: Any = None,
                 positive: Any = False):
        model = LinearRegression(fit_intercept=fit_intercept, normalize=normalize, copy_X=copy_X, n_jobs=n_jobs,
                                 positive=positive)

        super().__init__(model, inspect.currentframe())

    def __str__(self):
        return "SkLinearRegression"


class SkRidge(Regressor):
    """
    Class that implements the Ridge regressor from sklearn.
    The parameters one could pass are the same ones you would pass instantiating
    the regressor Ridge directly from sklearn.

    Sklearn documentation: [here](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.Ridge.html)

    """

    def __init__(self, alpha: Any = 1.0,
                 *,
                 fit_intercept: Any = True,
                 normalize: Any = "deprecated",
                 copy_X: Any = True,
                 max_iter: Any = None,
                 tol: Any = 1e-3,
                 solver: Any = "auto",
                 positive: Any = False,
                 random_state: Any = None):
        model = Ridge(alpha=alpha, fit_intercept=fit_intercept, normalize=normalize, copy_X=copy_X,
                      max_iter=max_iter, tol=tol, solver=solver, positive=positive, random_state=random_state)

        super().__init__(model, inspect.currentframe())

    def fit(self, X: Union[np.ndarray, sparse.csr_matrix], Y: list = None):
        self.model.fit(X.toarray() if isinstance(X, sparse.csr_matrix) else X, Y)

    def predict(self, X_pred: Union[np.ndarray, sparse.csr_matrix]):
        return self.model.predict(X_pred.toarray() if isinstance(X_pred, sparse.csr_matrix) else X_pred)

    def __str__(self):
        return "SkRidge"


class SkBayesianRidge(Regressor):
    """
    Class that implements the BayesianRidge regressor from sklearn.
    The parameters one could pass are the same ones you would pass instantiating
    the regressor BayesianRidge directly from sklearn.

    Sklearn documentation: [here](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.BayesianRidge.html)

    """

    def __init__(self, *,
                 n_iter: Any = 300,
                 tol: Any = 1.0e-3,
                 alpha_1: Any = 1.0e-6,
                 alpha_2: Any = 1.0e-6,
                 lambda_1: Any = 1.0e-6,
                 lambda_2: Any = 1.0e-6,
                 alpha_init: Any = None,
                 lambda_init: Any = None,
                 compute_score: Any = False,
                 fit_intercept: Any = True,
                 normalize: Any = "deprecated",
                 copy_X: Any = True,
                 verbose: Any = False):
        model = BayesianRidge(n_iter=n_iter, tol=tol, alpha_1=alpha_1, alpha_2=alpha_2, lambda_1=lambda_1,
                              lambda_2=lambda_2, alpha_init=alpha_init, lambda_init=lambda_init,
                              compute_score=compute_score, fit_intercept=fit_intercept, normalize=normalize,
                              copy_X=copy_X, verbose=verbose)

        super().__init__(model, inspect.currentframe())

    def fit(self, X: Union[np.ndarray, sparse.csr_matrix], Y: list = None):
        self.model.fit(X.toarray() if isinstance(X, sparse.csr_matrix) else X, Y)

    def predict(self, X_pred: Union[np.ndarray, sparse.csr_matrix]):
        return self.model.predict(X_pred.toarray() if isinstance(X_pred, sparse.csr_matrix) else X_pred)

    def __str__(self):
        return "SkBayesianRidge"


class SkSGDRegressor(Regressor):
    """
    Class that implements the SGD regressor from sklearn.
    The parameters one could pass are the same ones you would pass instantiating
    the regressor SGD directly from sklearn.

    Sklearn documentation: [here](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.SGDRegressor.html)
    """

    def __init__(self, loss: Any = "squared_error",
                 *,
                 penalty: Any = "l2",
                 alpha: Any = 0.0001,
                 l1_ratio: Any = 0.15,
                 fit_intercept: Any = True,
                 max_iter: Any = 1000,
                 tol: Any = 1e-3,
                 shuffle: Any = True,
                 verbose: Any = 0,
                 epsilon: Any = DEFAULT_EPSILON,
                 random_state: Any = None,
                 learning_rate: Any = "invscaling",
                 eta0: Any = 0.01,
                 power_t: Any = 0.25,
                 early_stopping: Any = False,
                 validation_fraction: Any = 0.1,
                 n_iter_no_change: Any = 5,
                 warm_start: Any = False,
                 average: Any = False):
        model = SGDRegressor(loss=loss, penalty=penalty, alpha=alpha, l1_ratio=l1_ratio, fit_intercept=fit_intercept,
                             max_iter=max_iter, tol=tol, shuffle=shuffle, verbose=verbose, epsilon=epsilon,
                             random_state=random_state, learning_rate=learning_rate, eta0=eta0, power_t=power_t,
                             early_stopping=early_stopping, validation_fraction=validation_fraction,
                             n_iter_no_change=n_iter_no_change, warm_start=warm_start, average=average)
        super().__init__(model, inspect.currentframe())

    def __str__(self):
        return "SkSGDRegressor"


class SkARDRegression(Regressor):
    """
    Class that implements the ARD regressor from sklearn.
    The parameters one could pass are the same ones you would pass instantiating
    the regressor ARD directly from sklearn.

    Sklearn documentation: [here](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.ARDRegression.html)
    """

    def __init__(self, *,
                 n_iter: Any = 300,
                 tol: Any = 1.0e-3,
                 alpha_1: Any = 1.0e-6,
                 alpha_2: Any = 1.0e-6,
                 lambda_1: Any = 1.0e-6,
                 lambda_2: Any = 1.0e-6,
                 compute_score: Any = False,
                 threshold_lambda: Any = 1.0e4,
                 fit_intercept: Any = True,
                 normalize: Any = "deprecated",
                 copy_X: Any = True,
                 verbose: Any = False):
        model = ARDRegression(n_iter=n_iter, tol=tol, alpha_1=alpha_1, alpha_2=alpha_2, lambda_1=lambda_1,
                              lambda_2=lambda_2, compute_score=compute_score, threshold_lambda=threshold_lambda,
                              fit_intercept=fit_intercept, normalize=normalize, copy_X=copy_X, verbose=verbose)
        super().__init__(model, inspect.currentframe())

    def fit(self, X: Union[np.ndarray, sparse.csr_matrix], Y: list = None):
        self.model.fit(X.toarray() if isinstance(X, sparse.csr_matrix) else X, Y)

    def predict(self, X_pred: Union[np.ndarray, sparse.csr_matrix]):
        return self.model.predict(X_pred.toarray() if isinstance(X_pred, sparse.csr_matrix) else X_pred)

    def __str__(self):
        return "SkARDRegression"


class SkHuberRegressor(Regressor):
    """
    Class that implements the Huber regressor from sklearn.
    The parameters one could pass are the same ones you would pass instantiating
    the regressor Huber directly from sklearn.

    Sklearn documentation: [here](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.HuberRegressor.html)
    """

    def __init__(self, *,
                 epsilon: Any = 1.35,
                 max_iter: Any = 100,
                 alpha: Any = 0.0001,
                 warm_start: Any = False,
                 fit_intercept: Any = True,
                 tol: Any = 1e-05):
        model = HuberRegressor(epsilon=epsilon, max_iter=max_iter, alpha=alpha,
                               warm_start=warm_start, fit_intercept=fit_intercept, tol=tol)

        super().__init__(model, inspect.currentframe())

    def __str__(self):
        return "SkHuberRegressor"


class SkPassiveAggressiveRegressor(Regressor):
    """
    Class that implements the PassiveAggressive regressor from sklearn.
    The parameters one could pass are the same ones you would pass instantiating
    the regressor PassiveAggressive directly from sklearn.

    Sklearn documentation: [here](https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.PassiveAggressiveRegressor.html)
    """

    def __init__(self, *,
                 C: Any = 1.0,
                 fit_intercept: Any = True,
                 max_iter: Any = 1000,
                 tol: Any = 1e-3,
                 early_stopping: Any = False,
                 validation_fraction: Any = 0.1,
                 n_iter_no_change: Any = 5,
                 shuffle: Any = True,
                 verbose: Any = 0,
                 loss: Any = "epsilon_insensitive",
                 epsilon: Any = DEFAULT_EPSILON,
                 random_state: Any = None,
                 warm_start: Any = False,
                 average: Any = False):

        model = PassiveAggressiveRegressor(C=C, fit_intercept=fit_intercept, max_iter=max_iter, tol=tol,
                                           early_stopping=early_stopping, validation_fraction=validation_fraction,
                                           n_iter_no_change=n_iter_no_change, shuffle=shuffle, verbose=verbose,
                                           loss=loss, epsilon=epsilon, random_state=random_state, warm_start=warm_start,
                                           average=average)
        super().__init__(model, inspect.currentframe())

    def __str__(self):
        return "SkPassiveAggressiveRegressor"



from .doc2vec import GensimDoc2Vec
from .fasttext import GensimFastText
from .latent_semantic_analysis import GensimLatentSemanticAnalysis
from .random_indexing import GensimRandomIndexing
from .word2vec import GensimWord2Vec
from .lda import GensimLDA



import os


def get_valid_filename(output_directory: str, filename: str, extension: str, overwrite: bool):
    """
    Method which gets a valid `filename.extension` based on the overwrite parameter

    If `overwrite=False` and there are existent file named as `filename.extension` in the output directory, the method
    checks if there are also files named as `filename (1).extension`, `filename (2).extension`,
    `filename (3).extension`, etc.
    It stops at the first non-existent `filename (x).extension` in the output directory specified
    and will return `filename (x).extension` as string

    Args:
        output_directory: Directory where the file will be saved
        filename: Name of the file to save
        extension: Extension of the file to save
        overwrite: Specify if the file to save should overwrite another file with
            same name if present. If `True` then simply `filename.extension` will be returned

    Returns:
        A valid `filename.extension` string based on the overwrite parameter
    """
    filename_try = "{}.{}".format(filename, extension)

    if overwrite is False:
        i = 0
        while os.path.isfile(os.path.join(output_directory, filename_try)):
            i += 1
            filename_try = "{} ({}).{}".format(filename, i, extension)

    return filename_try


def get_valid_dirname(output_directory: str,
                      directory_to_save: str,
                      overwrite: bool,
                      start_from_1: bool = False,
                      style: str = "parenthesis"):
    """
    Method which gets a valid directory name depending on the overwrite parameter.

    If `overwrite=False` and there is a directory named as `directory_to_save` in the `output_directory`, the method
    checks if there are also directories named as `directory_to_save (1)`, `directory_to_save (2)`,
    `directory_to_save (3)`. etc.
    It stops at the first non-existent `directory_to_save (x)` in the `output_directory` specified
    and will return `directory_to_save (x)` as string

    There are two styles supported:

    * `parenthesis` (default) -> will produce strings of the type 'directory_to_save (1)', directory_to_save (2), etc.
    * `underscore` -> will produce strings of the type 'directory_to_save_1', directory_to_save_2, etc.


    Args:
        output_directory: Directory where the `directory_to_save` will be created
        directory_to_save: Name of the directory to save
        overwrite: Specifies if the directory to save should overwrite another directory with
            same name if present.
        start_from_1: Specifies if, regardless of any other parameter, the string to return should contain (1) or '_1'
            depending on the `style` parameter chosen even if there are no other directories with same name
        style: can be `parenthesis` or `underscore`:

            * `parenthesis` (default) -> will produce strings of the type 'directory_to_save (1)',
                directory_to_save (2), etc.
            * `underscore` -> will produce strings of the type 'directory_to_save_1',
                directory_to_save_2, etc.

    Returns:
        A valid directory name

    Raises:
        `ValueError` exception if `style` different from 'parenthesis' or 'underscore'
    """
    valid_styles = {'parenthesis', 'underscore'}
    if style.lower() not in valid_styles:
        raise ValueError(f"Style {style} not supported! Only {valid_styles} are supported")

    dirname_try = directory_to_save
    if start_from_1 and style == "underscore":
        dirname_try = f"{directory_to_save}_1"
    elif start_from_1 and style == "parenthesis":
        dirname_try = f"{directory_to_save} (1)"

    if overwrite is False:
        i = 0
        while os.path.isdir(os.path.join(output_directory, dirname_try)):
            i += 1
            if style == "underscore":
                dirname_try = "{}_{}".format(directory_to_save, i)
            else:
                dirname_try = "{} ({})".format(directory_to_save, i)

    return dirname_try



from abc import ABC, abstractmethod
from typing import Tuple
import numpy as np


class ScoreProcessor(ABC):
    """
    Abstract class to process a rating with the personalized fit method
    """
    def __init__(self, decimal_rounding: int = None):
        self.__decimal_rounding = decimal_rounding

    @property
    def decimal_rounding(self):
        return self.__decimal_rounding

    @abstractmethod
    def fit(self, score_data: object):
        raise NotImplementedError

    def __repr__(self):
        return f'ScoreProcessor(decimal rounding={self.__decimal_rounding})'


class SentimentAnalysis(ScoreProcessor):
    """
    Abstract Class that generalizes the sentiment analysis technique
    """

    @abstractmethod
    def fit(self, score_data: str):
        raise NotImplementedError

    @abstractmethod
    def __repr__(self):
        return f'SentimentAnalysis(decimal rounding={self.decimal_rounding})'


class NumberNormalizer(ScoreProcessor):
    """
    Class that normalizes numeric scores to a scale in the range $[-1.0, 1.0]$

    Args:
        scale: Tuple where the first value is the minimum of the actual scale, second value is the maximum of the
            actual scale (e.g. `(1, 5)` represents an actual scale of scores from 1 (included) to 5 (included))
        decimal_rounding: If set, the normalized score will be rounded to the chosen decimal digit
    """
    def __init__(self, scale: Tuple[float, float], decimal_rounding: int = None):
        super().__init__(decimal_rounding)

        if len(scale) != 2:
            raise ValueError("The voting scale should be a tuple containing exactly two values,"
                             "the minimum of the scale and the maximum!")

        self._old_min = scale[0]
        self._old_max = scale[1]

    def __str__(self):
        return "NumberNormalizer"

    def __repr__(self):
        return f'NumberNormalizer(scale=({self._old_min}, {self._old_max}), decimal rounding={self.decimal_rounding})'

    def fit(self, score_data: float) -> float:
        """
        Method which will normalize the given score

        Args:
            score_data: score that will be normalized

        Returns:
            score normalized in the interval $[-1, 1]$
        """
        def convert_into_range(value: float, old_min: float, old_max: float, new_min: int = -1, new_max: int = 1):
            new_value = ((value - old_min) / (old_max - old_min)) * (new_max - new_min) + new_min
            if self.decimal_rounding:
                new_value = np.round(new_value, self.decimal_rounding)

            return new_value

        return convert_into_range(float(score_data), self._old_min, self._old_max)
